""" Bid for Game
    Acol OpenersLaterBid module
"""

import inspect
from .bidding import Bid, Pass
from .blackwood import Blackwood
from .hand import Hand

inspection = inspect.currentframe


class OpenersLaterBid(Hand):
    """BfG OpenersLaterBid class."""
    def __init__(self, hand_cards, board):
        super(OpenersLaterBid, self).__init__(hand_cards, board)
        self.my_first_bid = Bid(self.bid_history[0], '')
        self.my_second_bid = Bid(self.bid_history[4], '')
        self.my_last_bid = Bid(self.bid_history[-4], '')
        self.my_penultimate_bid = Bid(self.bid_history[-8], '')
        self.partners_last_bid = Bid(self.bid_history[-2], '')
        self.partners_penultimate_bid = Bid(self.bid_history[-6], '')
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and returns a Bid object."""
        if self.my_last_bid.name == '4NT':
            bid = self.bid_after_4nt()
        elif self.my_last_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).select_contract()
        elif self.partners_last_bid.name == '4NT':
            bid = Blackwood(self.cards, self.board).count_aces()
        elif self.partners_last_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).count_kings()
        elif (self.hcp >= 23 and
                self.responder_bid_two.denomination.is_suit and
                self.responder_bid_two.denomination != self.responder_bid_one.denomination and
                self.next_level(self.responder_bid_two.denomination) < 6):
            bid = self._strong_and_responder_changes_suit()
        elif (self.responder_bid_one.denomination != self.responder_bid_two.denomination and
              self.responder_bid_three.denomination == self.opener_suit_one and
              self.responder_bid_one.denomination.is_major and
              self.suit_length(self.responder_bid_one.denomination) >= 3 and
              not Bid(self.bid_history[-2]).is_game):
            bid = self.bid_to_game(self.responder_bid_one.denomination, '1801')
        elif self._responder_supports_last_bid_suit():
            bid = self.bid_to_game(self.partners_last_bid.denomination, '1802')
        # elif self._responder_supports_first_bid_suit():
        #     bid = self.bid_to_game(self.partners_last_bid.denomination, '1809')
        elif self.partners_last_bid.level >= 6:
            bid = Pass('1810')
        elif (self.partners_last_bid.is_game or
                (self.partners_penultimate_bid.is_game and self.my_last_bid.is_game)):
            bid = Pass('1811')
        elif self.partners_last_bid.denomination == self.my_first_bid.denomination:
            bid = self.next_level_bid(self.my_first_bid.denomination, '1812')
        # NB same logic as above
        # elif self.partners_last_bid.denomination == self.my_second_bid.denomination:
        #     bid = self.next_level_bid(self.my_second_bid.denomination, '1812')
        elif self._partner_has_invited_game_minor() and self.hcp >= 14:
            bid = self.bid_to_game(self.partner_last_bid.denomination, '1816')
        elif self._opponents_at_game():
            bid = Pass('1813')
        elif self._partner_has_invited_game_minor() and self.hcp <= 13:
            bid = Pass('1814')
        # elif self.partner_bid_two.is_game:
        #     bid = Pass('1819')
        elif self._opponents_have_doubled() and self.partner_last_bid.is_pass:
            bid = Pass('1817')
        elif ((self.partner_bid_one.is_pass or self.partner_bid_two.is_pass) and
                self.partner_last_bid.is_pass):
            bid = Pass('1815')
        elif self.partner_last_bid.is_pass and self.hcp <= 14:
            bid = Pass('1818')
        else:
            bid = Pass('1803')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def bid_after_4nt(self):
        """Return bid after response to 4NT."""
        four_aces = Blackwood(self.cards, self.board).four_aces
        if four_aces and self.kings != 4:
            bid = self.nt_bid(5, '1804')
        elif (Blackwood(self.cards, self.board).has_three_aces and
              self.kings == 4 and self._suit_fit()):
            bid = self.suit_bid(6, self._suit_fit(), '1805')
        else:
            bid = Blackwood(self.cards, self.board).select_contract()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _strong_and_responder_changes_suit(self):
        """Return bid after responder changes suit, showing 5/5."""
        if self.suit_length(self.responder_bid_one.denomination) >= 3:
            bid = self.suit_bid(6, self.responder_bid_one.denomination, '1806')
        elif self.suit_length(self.responder_bid_two.denomination) >= 3:
            bid = self.suit_bid(6, self.responder_bid_two.denomination, '1807')
        else:
            bid = self.nt_bid(6, '1808')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    # Various utility functions
    def _suit_fit(self):
        """Return the suit if a fit is found."""
        suits = []
        our_bids = [bid for bid in self.bid_history[::-1][1::2]]
        responders_bids = our_bids[::2]
        for bid in responders_bids:
            call = Bid(bid)
            if call.level <= 4 and call.is_suit_call:
                if call.denomination != self.club_suit:
                    suits.append(call.denomination)
                elif call.denomination == self.club_suit and Bid(self.bid_history[0]).is_suit_call:
                    suits.append(call.denomination)
        for suit in suits:
            if self.suit_length(suit) >= 4:
                return suit
        return None

    # Various boolean functions

    def _responder_supports_last_bid_suit(self):
        """Return True if responder supports last bid suit."""
        result = (self.my_last_bid.denomination != self.my_penultimate_bid.denomination and
                  (self.partners_last_bid.denomination == self.my_last_bid.denomination or
                   self.partners_last_bid.denomination == self.my_penultimate_bid.denomination) and
                  self.partners_last_bid.denomination.is_suit and
                  not self.bidding_above_game)
        return result

    def _responder_supports_first_bid_suit(self):
        """Return True if responder supports fisrt bid suit."""
        result = (self.my_last_bid.denomination != self.my_penultimate_bid.denomination and
                  (self.partners_last_bid.denomination == self.my_first_bid.denomination or
                   self.partners_last_bid.denomination == self.my_penultimate_bid.denomination) and
                  self.partners_last_bid.denomination.is_suit and
                  not self.bidding_above_game)
        return result

    def _partner_has_invited_game_minor(self):
        """Return True if partner's bid 4 minor."""
        result = (self.partners_last_bid.is_minor and
                  self.partner_last_bid.level == 4)
        return result



""" Bid for Game
    Acol Opener's third bid module
"""

import inspect
from .bidding import Pass
from .blackwood import Blackwood
from .hand import Hand

inspection = inspect.currentframe


class OpenersThirdBid(Hand):
    """BfG OpeningThirdBid class."""
    def __init__(self, hand_cards, board):
        super(OpenersThirdBid, self).__init__(hand_cards, board)
        self.last_level = 1
        self.responder_weak_1nt = (self.responder_bid_one.name == '1NT')
        self.opener_two_suited = self.bid_one.denomination != self.bid_two.denomination
        self.stayman_bid = (self.bid_one.is_nt and
                            self.responder_bid_one.denomination == self.club_suit and
                            self.last_bid.is_pass)
        self.trace = 0

    def suggested_bid(self):
        """Directs control to relevant method and returns a Bid object."""
        if self.bid_two.name == '4NT':
            bid = self._blackwood_bids()
        elif self.responder_bid_two.name == '4NT':
            bid = Blackwood(self.cards, self.board).count_aces()
        elif self.bid_two.is_pass or self.responder_bid_two.is_pass:
            bid = self._weak_contested_auction()
        elif self.responder_bid_two.is_nt:
            bid = self._responder_bids_nt()
        elif self._four_suits_bid():
            bid = self._rebid_after_four_suits()
        elif self._strong_responder_better_than_minimum():
            bid = self.nt_bid(4, '1601')
        elif self._both_partners_better_than_minimum_and_no_overcall():
            bid = self.nt_bid(4, '1602')
        elif self.responder_bid_two.is_game:
            bid = Pass('1603')
        elif self.last_level >= 5:
            bid = Pass('1723')
        elif self._responder_supports_opener():
            bid = self._responder_support()
        elif self._three_card_support_for_responders_major():
            bid = self.next_level_bid(self.responder_bid_one.denomination, '1604')
        elif (self.responder_bid_one.denomination == self.responder_bid_two.denomination and
              not self.responder_bid_one.is_pass):
            bid = self._responder_repeats_suit()
        elif self.five_five_or_better:
            bid = self._rebid_with_five_five()
        elif self._has_seven_card_major():
            bid = self.suit_bid(4, self.longest_suit, '1605')
        elif self._responder_bid_one_is_clubs():
            bid = self._support_for_responder()
        elif self._three_card_support_for_responders_second_suit():
            bid = self.next_level_bid(self.responder_bid_two.denomination, '1606')
        elif self.overcaller_bid_one.is_double and self.shape[0] >= 6:
            bid = self.next_level_bid(self.longest_suit, '1607')
        elif self._three_suits_bid():
            bid = self.next_nt_bid('1608')
        elif self.shape[0] >= 6:
            bid = self._rebid_long_suit()
        elif self.bid_two.name == '2NT' and self.responder_bid_two.denomination.is_major:
            bid = self._responder_shows_five_card_major()
        else:
            bid = Pass('1609')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_shows_five_card_major(self):
        """Return bid after 2NT when responder shows a 5 card suit."""
        suit = self.responder_bid_two.denomination
        if self.suit_length(suit) >= 3:
            bid = self.bid_to_game(suit, '1610')
        elif self.nt_level <= 3:
            bid = self.bid_to_game(self.no_trumps, '1611')
        elif (self.overcallers_last_bid.level >= 4 and
              self.suit_holding[self.responder_bid_two.denomination] <= 2):
            bid = Pass('1612')
        else:
            bid = Pass('1724')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _blackwood_bids(self):
        """Bid if in Blackwood process."""
        if Blackwood(self.cards, self.board).four_aces:
            if self.responder_bid_one.level == 4 and self.hcp <= 22:
                bid = self.suit_bid(6, self.bid_one.denomination, '1613')
            else:
                bid = self.nt_bid(5, '1614')
        else:
            bid = Blackwood(self.cards, self.board).select_contract()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_contested_auction(self):
        """Bid if contested but weak or partner passes."""
        preference_suit = self._show_preference()
        if self._four_card_support_for_responder_and_passed():
            bid = self.next_level_bid(self.responder_bid_one.denomination, '1615')
        elif self.overcaller_bid_two.is_game:
            bid = Pass('1719')
        elif (self.overcaller_bid_one.is_double and
              self.responder_bid_two.is_pass and
              not (self.overcaller_bid_two.is_pass and
                   self.advancer_bid_two.is_pass)):
            bid = Pass('1616')
        elif self.shape[0] >= 7:
            bid = self._rebid_long_suit()
        elif self._can_rebid_long_suit_at_level_three():
            bid = self.next_level_bid(self.longest_suit, '1617')
        elif self._three_suits_bid_contested_auction(preference_suit):
            bid = self.bid_to_game(preference_suit, '1618')
        elif (self.hcp >= 16 and
                self.shape[0] >= 6 and
                self.next_level(self.bid_one.denomination) <= 3):
            bid = self.next_level_bid(self.bid_one.denomination, '1619')
        elif self._can_support_responders_four_card_major():
            bid = self.next_level_bid(self.responder_bid_two.denomination, '1620')
        elif self._responder_promises_stopper():
            bid = self.nt_bid(3, '1621')
        elif (self.hcp <= 14 and
              self.bid_one.is_minor and
              self.responder_bid_two.denomination == self.bid_one.denomination):
            bid = Pass('1712')
        elif self.next_level(self.bid_one.denomination) > 4:
            bid = Pass('1713')
        elif self.hcp <= 10:
            bid = Pass('1714')
        elif self.bid_one.name == '1NT' and self.bid_two.is_pass:
            bid = Pass('1715')
        elif self.responder_bid_two.is_pass:
            bid = Pass('1716')
        else:
            bid = Pass('1622')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_support(self):
        """Rebid with responder support."""
        if self._responder_has_jumped_in_minor():
            bid = self.nt_bid(3, '1623')
        elif self._intermediate_with_six_four():
            bid = self.next_level_bid(self.opener_suit_one, '1624')
        elif self._responder_repeats_suit_and_can_support():
            bid = self.next_level_bid(self.opener_suit_two, '1625')
        elif self._responder_has_invited_in_six_card_suit():
            bid = self.bid_to_game(self.longest_suit, '1626')
        elif self._responder_has_invited_in_second_suit():
            bid = self.bid_to_game(self.second_suit, '1705')
        elif self._weak_and_two_suited():
            bid = Pass('1628')
        elif self.opener_bid_one.name == '1NT' and self.hcp == 14:
            bid = self.bid_to_game(self.opener_suit_two, '1629')
        elif not self.responder_weak_1nt or self.hcp >= 17:
            bid = self._responder_support_strong()
        elif self.responder_bid_one.level >= 2:
            bid = self._responder_support_strong()
        else:
            bid = Pass('1630')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_support_strong(self):
        """Rebid with responder support with strong hand."""
        if (self.responder_bid_one.is_pass and
                self.responder_bid_two.denomination == self.opener_suit_one and
                self.responder_bid_two.level == 2):
            bid = Pass('1721')
        elif (self.responder_bid_two.denomination == self.opener_suit_one and
                self.responder_bid_two.level == 2):
            bid = Pass('1631')
        elif self._can_bid_again():
            bid = self._responder_support_strong_constructive()
        elif self._better_than_minimum_six_card_major():
            bid = self.bid_to_game(self.longest_suit, '1632')
        elif self._respsonder_invites_game_in_minor():
            bid = self.next_level_bid(self.opener_suit_one, '1633')
        else:
            bid = Pass('1634')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_support_strong_constructive(self):
        """Rebid with responder support - constructive"""
        if self._responder_rebids_minor_and_semi_balanced():
            bid = self.nt_bid(3, '1635')
        elif self.hcp >= 19 and self.is_balanced and self.nt_level <= 3:
            bid = self.nt_bid(3, '1636')
        elif self._strong_and_responder_supports_four_card_major():
            bid = self.nt_bid(3, '1637')
        elif self._strong_or_responder_bids_minor_at_level_four():
            bid = self.next_level_bid(self.responder_bid_two.denomination, '1638')
        elif self._responder_has_jumped_in_openers_minor():
            bid = self.bid_to_game(self.opener_suit_one, '1639')
        elif self.responder_bid_two.denomination.is_minor:
            bid = Pass('1640')
        elif self.hcp >= 15 and self.responder_bid_one.level >= 2:
            if (self.responder_bid_two.denomination == self.bid_two.denomination and
                    self.hcp >= 15):
                bid = self.next_level_bid(self.bid_two.denomination, '1641')
            else:
                bid = self.nt_bid(3, '1642')
        elif self._has_five_card_major() and not self.responder_weak_bid():
            bid = self.suit_bid(4, self.bid_one.denomination, '1643')
        elif self._responder_has_jumped_in_openers_major():
            suit = self.opener_suit_one
            if (self.suit_length(suit) <= 4 and
                    self.stoppers_in_bid_suits and self.nt_level <= 3):
                bid = self.nt_bid(3, '1646')
            else:
                bid = self.bid_to_game(suit, '1646')
        elif (16 <= self.hcp <= 17 and
              self.responder_bid_two.denomination == self.bid_two.denomination):
            bid = Pass('1647')
        elif (self.responder_bid_one.is_pass):
            bid = Pass('1717')
        else:
            bid = Pass('1718')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _support_for_responder(self):
        """Rebid with support for responder"""
        suit = self._select_better_suit_from_partner()
        level = self.next_level(suit)
        opener_previously_jumped = False  # see hand 3_9_9
        if self.hcp >= 16 and not opener_previously_jumped and level < 4:
            level += 1
        if (suit.is_minor and
                self.three_suits_bid_and_stopper() and
                self.nt_level <= 3):
            bid = self.nt_bid(3, '1648')
        elif self.shape[0] >= 6 and self.longest_suit.is_major:
            bid = self.next_level_bid(self.longest_suit, '1649')
        elif (self.is_jump(self.opener_bid_two, self.responder_bid_two) and
                self.responder_bid_one.denomination.is_major and
                self.suit_length(self.responder_bid_one.denomination) >= 3 and
                self.next_level(self.responder_bid_one.denomination) <= 4):
            bid = self.bid_to_game(self.responder_bid_one.denomination, '')
        elif ((suit == self.responder_bid_one.denomination and
               self.suit_holding[suit] >= 3) or
              (suit == self.responder_bid_two.denomination and
               self.suit_holding[suit] >= 4)):
            bid = self.suit_bid(level, suit, '1650')
        else:
            bid = self.suit_bid(level, suit, '1716')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_bids_nt(self):
        """Rebid after responder bids NT"""
        if self.responder_bid_two.name == '6NT':
            bid = Pass('1651')
        elif self.responder_bid_two.name == '5NT' and self.bid_history[-6] == '4NT':
            bid = Blackwood(self.cards, self.board).count_kings()
        elif self.responder_bid_two.name == '2NT':
            bid = self._responder_bids_two_nt()
        elif self.hcp == 19 and self.responder_bid_two.name == '3NT':
            bid = self.nt_bid(6, '1652')
        elif self._responder_bids_three_nt_after_stayman_two_four_card_majors():
            bid = self._responder_bids_3nt_and_stayman()
        elif self.hcp >= 16 and self.five_five_or_better:
            bid = self.next_level_bid(self.bid_two.denomination, '1653')
        elif (self.five_five_or_better and
                self.second_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.second_suit, '1654')
        elif self.hcp >= 16 and self.nt_level <= 3:
            bid = self.next_nt_bid('1655')
        elif self.responder_bid_two.is_game:
            bid = Pass('1656')
        else:
            bid = Pass('1728')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_bids_two_nt(self):
        """Rebid after responder bids 2NT"""
        if self._better_than_minimum_and_five_card_suit():
            bid = self.nt_bid(3, '1657')
        elif self._can_call_3nt():
            bid = self.nt_bid(3, '1658')
        elif self.shape[0] >= 7:
            bid = self.next_level_bid(self.longest_suit, '1659')
        elif self.stayman_bid:
            bid = self._nt_has_been_bid_after_stayman()
        elif (self.hcp >= 15 and
              self.opener_bid_two.is_suit_call and
              self.responder_bid_one.is_suit_call and
              (self.shape[0] <= 5 or
               self.longest_suit.is_minor)):
            bid = self.next_nt_bid('1660')
        elif self.shape[0] >= 6 and self.hcp <= 14:
            bid = self.next_level_bid(self.longest_suit, '1661')
        elif self.shape[0] >= 6:
            bid = self.bid_to_game(self.longest_suit, '1662')
        else:
            bid = Pass('1663')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _nt_has_been_bid_after_stayman(self):
        """Return bid if NT has been bid after Stayman."""
        suit = None
        if self._two_four_card_majors_after_stayman():
            if (self.bid_two.denomination == self.heart_suit and
                    self.spade_suit not in self.opponents_suits):
                suit = self.spade_suit
            elif self.heart_suit not in self.opponents_suits:
                suit = self.heart_suit
            if suit:
                if self.next_level(suit) <= 4:
                    bid = self.bid_to_game(suit, '1664')
                # elif self.nt_level <= 3:
                #     bid = self.nt_bid(3, '1665')
                # else:
                #     bid = Pass('1666')
            elif self.hcp >= 13 and self.nt_level <= 3:
                bid = self.nt_bid(3, '1711')
            else:
                bid = Pass('1708')
        elif self._maximum_after_stayman():
            bid = self.nt_bid(3, '1668')
        elif self._minimum_after_stayman():
            bid = Pass('1669')
        else:
            bid = Pass('1725')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_bids_3nt_and_stayman(self):
        """Rebid after responder bids 3NT and stayman"""
        if self._has_two_four_card_majors_after_hearts():
            bid = self.next_level_bid(self.spade_suit, '1670')
        elif self._has_two_four_card_majors_after_spades():
            bid = self.next_level_bid(self.heart_suit, '1671')
        elif self.responder_bid_two.name == '3NT':
            bid = Pass('1709')
        else:
            bid = Pass('1726')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _rebid_long_suit(self):
        """Rebid own long suit"""
        if self.responder_bid_two.is_game:
            bid = Pass('1673')
        elif self.shape[0] >= 8:
            bid = self._rebid_long_suit_eight_cards()
        elif self.shape[0] == 7:
            bid = self._rebid_long_suit_seven_cards()
        elif self.shape[0] == 6:
            if self._responder_has_jumped_in_major():
                bid = self.next_level_bid(self.responder_bid_one.denomination, '1674')
            else:
                bid = self._rebid_long_suit_six_cards()
        else:
            bid = Pass('1710')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _rebid_long_suit_eight_cards(self):
        """Rebid own long suit - 8+ cards"""
        suit = self.longest_suit
        if self.next_level(suit) <= 4:
            bid = self.suit_bid(4, suit, '1676')
        else:
            bid = Pass('1677')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _rebid_long_suit_seven_cards(self):
        """Rebid own long suit - 7 cards"""
        if self.shape[1] >= 5 and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '1678')
        elif self._can_bid_strong_major_at_level_three():
            bid = self.next_level_bid(self.longest_suit, '1679')
        elif self.shape[0] >= 7 and self.hcp >= 15:
            bid = self.next_level_bid(self.longest_suit, '1680')
        elif (self.hcp <= 14 and self.responder_bid_one.name == '1NT' and
              not self.overcaller_bid_one.is_pass):
            bid = Pass('1681')
        else:
            bid = Pass('1722')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _rebid_long_suit_six_cards(self):
        """Rebid own long suit - 6 cards"""
        if self.shape[1] >= 5 and self.second_suit not in self.opponents_suits:
            suit = self.second_suit
            bid = self.next_level_bid(suit, '1682')
        elif (self.is_jump(self.bid_one,  self.responder_bid_one) and self.hcp >= 12):
            bid = self.next_level_bid(self.longest_suit, '1683')
        elif self.responder_bid_one.level == 2 and self.hcp >= 15:
            bid = self.next_level_bid(self.longest_suit, '1684')
        elif self._responder_changes_suit_at_level_three():
            bid = self.next_level_bid(self.longest_suit, '1685')
        elif (self.partner_bid_one.is_suit_call and
                self.partner_bid_one.denomination != self.partner_bid_two.denomination):
            bid = self.next_level_bid(self.longest_suit, '1686')
        # elif self._responder_jumps_in_major():
        #     if self.suit_length(self.partner_bid_one.denomination) >= 2:
        #         bid = self.bid_to_game(self.partner_bid_one.denomination, '1687')
        #     else:
        #         bid = self.nt_bid(3, '1688')
        elif (self.hcp <= 15 and
              self.responder_bid_one.denomination == self.responder_bid_two.denomination):
            bid = Pass('1689')
        else:
            bid = Pass('1720')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_repeats_suit(self):
        """Rebid responder has repeated suit"""
        if self._is_unbalanced_and_has_two_card_support_for_responders_major():
            bid = self.next_level_bid(self.responder_bid_one.denomination, '1690')
        elif self.shape[0] >= 6:
            bid = self._rebid_long_suit()
        elif self.is_balanced and self.hcp >= 18 and self.nt_level <= 3:
            bid = self.next_nt_bid('1691')
        elif self.is_balanced and self.responder_bid_two.level <= 3:
            bid = Pass('1692')
        else:
            bid = self._support_responders_rebid_suit()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _support_responders_rebid_suit(self):
        """Rebid responder has repeated suit, support suit."""
        responders_suit = self.responder_bid_one.denomination
        # hand_value_points = (self.hcp + self.support_shape_points(responders_suit))
        if self.five_five_or_better and self.hcp >= 13:
            bid = self._rebid_with_five_five()
        elif self.bid_one.name == '2C' and self.five_four_or_better:
            bid = self.next_level_bid(self.second_suit, '1693')
        elif self._can_raise_responders_invitation():
            bid = self.next_level_bid(responders_suit, '1694')
        elif self._strong_with_stoppers():
            bid = self.nt_bid(3, '1698')
        elif self._responder_repeats_suit_after_two_nt():
            bid = self.nt_bid(3, '1699')
        elif self._partner_bid_minor_at_level_four_two_card_support():
            bid = self.suit_bid(6, responders_suit, '1700')
        elif self.suit_holding[self.responder_bid_two.denomination] <= 2:
            bid = Pass('1701')
        else:
            bid = Pass('1729')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _rebid_with_five_five(self):
        """Rebid with 5/5 hand."""
        suit = self.bid_two.denomination
        if (self.partner_bid_one.name == '1NT' and
                self.partner_bid_two.level == 3):
            bid = Pass('1702')
        elif self.suit_length(self.responder_bid_two.denomination) >= 5:
            bid = self.next_level_bid(self.responder_bid_two.denomination, '1703')
        elif True:
            bid = self.next_level_bid(suit, '1704')
        else:
            bid = self.next_level_bid(suit, '1727')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _rebid_after_four_suits(self):
        """Rebid after 4 suits bid."""
        suit = self._select_after_four_suits()
        if suit == self.responder_bid_two.denomination and self.responder_bid_two.is_game:
            bid = Pass('1705')
        elif (self.nt_level <= 3 and
                self.suit_stopper(self.responder_bid_two.denomination)):
            bid = self.nt_bid(3, '1706')
        else:
            bid = self.next_level_bid(suit, '1707')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    # Various utility functions

    def _select_better_suit_from_partner(self):
        """Return the best suit after parner shows two"""
        if (self.responder_bid_two.denomination.is_major and
                self.suit_length(self.responder_bid_two.denomination) >= 4):
            return self.responder_bid_two.denomination
        elif self.suit_length(self.responder_bid_one.denomination) >= 3:
            return self.responder_bid_one.denomination
        elif self.suit_length(self.responder_bid_two.denomination) >= 3:
            return self.responder_bid_two.denomination
        assert False, 'suit not assigned'

    def _select_after_four_suits(self):
        """Select suit after 4 have been bid."""
        suit_one = self.responder_bid_one.denomination
        suit_two = self.responder_bid_two.denomination
        holding_one = 6 + self.suit_length(suit_one)
        holding_two = 5 + self.suit_length(suit_two)
        if holding_one == holding_two:
            if suit_one.is_major:
                suit = suit_one
            elif suit_two.is_major:
                suit = suit_two
            else:
                suit = suit_one
        elif holding_one > holding_two:
            suit = suit_one
        else:
            suit = suit_two
        if self.suit_length(suit) <= 1:
            if self.shape[1] >= 5:
                suit = self.second_suit
            else:
                suit = self.longest_suit
        self.tracer(__name__, inspection(), suit, self.trace)
        return suit

    # Various boolean functions

    def _four_suits_bid(self):
        """Return True if all 4 suits have been bid."""
        value = False
        suits_bid = (self.bid_one.is_suit_call and
                     self.bid_two.is_suit_call and
                     self.responder_bid_one.is_suit_call and
                     self.responder_bid_two.is_suit_call)
        if (suits_bid and
                self.bid_one.denomination != self.bid_two.denomination and
                self.bid_one.denomination != self.responder_bid_one.denomination and
                self.bid_one.denomination != self.responder_bid_two.denomination and
                self.bid_two.denomination != self.responder_bid_one.denomination and
                self.bid_two.denomination != self.responder_bid_two.denomination and
                self.responder_bid_one.denomination != self.responder_bid_two.denomination):
            value = True
        return value

    def _responder_supports_opener(self):
        """Return True if responder supports opener."""
        value = False
        if (self.responder_bid_two.denomination == self.bid_one.denomination or
                self.responder_bid_two.denomination == self.bid_two.denomination):
            value = True
        return value

    def _show_preference(self):
        """Return suit preference."""
        if self.responder_bid_one.is_suit_call:
            holding_one = self.suit_length(self.responder_bid_one.denomination)
        else:
            holding_one = 0
        if self.responder_bid_two.is_suit_call:
            holding_two = self.suit_length(self.responder_bid_two.denomination)
        else:
            holding_two = 0
        if holding_one >= holding_two + 1:
            suit = self.responder_bid_one.denomination
        else:
            suit = self.responder_bid_two.denomination
        return suit

    def _strong_responder_better_than_minimum(self):
        """Return True if strong and responder better than minimum."""
        return (self.is_jump(self.bid_one, self.responder_bid_one) and
                not self.overcall_made and
                self.hcp >= 17 and
                self.nt_level <= 4)

    def _both_partners_better_than_minimum_and_no_overcall(self):
        """Return True if better than minimum and no overcall."""
        return (self.is_jump(self.bid_one, self.responder_bid_one) and
                not self.overcall_made and
                self.hcp >= 15 and
                self.nt_level <= 4 and
                not self.is_jump(self.bid_two, self.responder_bid_two))

    def _three_card_support_for_responders_major(self):
        """Return True if 3 card support for responders major."""
        return (self.responder_bid_one.denomination == self.responder_bid_two.denomination and
                self.responder_bid_one.denomination.is_major and
                self.suit_length(self.responder_bid_one.denomination) >= 3 and
                self.next_level(self.responder_bid_one.denomination) <= 4)

    def _has_seven_card_major(self):
        """Return True if hand has 7 card major."""
        return (self.shape[0] >= 7 and
                self.longest_suit.is_major and
                self.next_level(self.longest_suit) <= 4)

    def _responder_bid_one_is_clubs(self):
        """Return True if responder has bid clubs."""
        if self.responder_bid_one.is_suit_call:
            responder_suit_length_one = self.suit_length(self.responder_bid_one.denomination)
        else:
            responder_suit_length_one = 0
        if self.responder_bid_two.is_suit_call:
            responder_suit_length_two = self.suit_length(self.responder_bid_two.denomination)
        else:
            responder_suit_length_two = 0
        return ((responder_suit_length_one >= 3 or responder_suit_length_two >= 4) and
                self.responder_bid_one.is_suit_call and
                self.responder_bid_two.is_suit_call and
                not (self.bid_one.is_nt and
                self.responder_bid_one.denomination == self.club_suit))

    def _three_card_support_for_responders_second_suit(self):
        """Return True if support for responder's major."""
        return (self.bid_one.name == '1NT' and
                self.responder_bid_two.denomination.is_major and
                self.suit_length(self.responder_bid_two.denomination) >= 3)

    def _three_suits_bid(self):
        """Return True if 3 suits bid."""
        return (self.three_suits_bid_and_stopper() and
                self.responder_bid_one.name != '1NT' and
                self.overcaller_bid_two.name != '2NT' and
                self.nt_level <= 3)

    def _can_rebid_long_suit_at_level_three(self):
        """Return True if can bid long suit at 3 level."""
        return (self.suit_length(self.longest_suit) >= 5 and
                self.suit_points(self.longest_suit) >= 8 and
                self.responder_bid_one.is_value_call and
                self.longest_suit not in self.opponents_suits and
                self.next_level_bid(self.longest_suit).level <= 3 and
                # self.hcp >= 16 and
                self.longest_suit.is_major)

    def _three_suits_bid_contested_auction(self, preference_suit):
        """Return True if 3 suits bid and contested auction."""
        return (self.responder_bid_one.denomination != self.responder_bid_two.denomination and
                self.responder_bid_two.denomination != self.bid_one.denomination and
                self.responder_bid_two.level >= 3 and
                self.responder_bid_one.is_suit_call and
                self.responder_bid_two.is_suit_call and
                self.next_level(preference_suit) <= preference_suit.game_level)

    def _four_card_support_for_responder_and_passed(self):
        """Return True four card support for responder."""
        return (self.responder_bid_one.is_suit_call and
                self.bid_history[1] != 'P' and
                self.opener_bid_two.is_pass and
                self.suit_length(self.responder_bid_one.denomination) >= 4 and
                self.next_level(self.responder_bid_one.denomination) <= 3)

    def _responder_has_jumped_in_minor(self):
        """Return True responder has bid in a minor."""
        return (self.is_semi_balanced and
                self.hcp >= 15 and
                self.responder_bid_two.denomination.is_minor and
                self.is_jump(self.responder_bid_one, self.responder_bid_two) and
                self.stoppers_in_unbid_suits() and
                not self.overcall_made and
                self.nt_level <= 3)

    def _weak_and_two_suited(self):
        """Return True if weak and two suited."""
        return (self.opener_two_suited and
                self.hcp <= 15 and
                self.responder_bid_one.level == 1)

    def _can_bid_again(self):
        """Return True can bid again."""
        return ((self.hcp >= 14 and
                not self.responder_bid_one.is_pass) or
                (self.hcp >= 15 and
                self.responder_bid_one.level >= 2) or
                (self.hcp >= 17 and
                self.responder_bid_two.level >= 2) or
                self.hcp >= 18)

    def _better_than_minimum_six_card_major(self):
        """Return True if six card major."""
        return (self.shape[0] >= 6 and
                self.hcp >= 13 and
                self.longest_suit.is_major and
                self.next_level(self.longest_suit) <= 4)

    def _responder_rebids_minor_and_semi_balanced(self):
        """Return True if responder changes to a minor."""
        return (self.responder_bid_two.denomination.is_minor and
                self.responder_bid_two.denomination != self.responder_bid_one.denomination and
                self.next_level_bid(self.responder_bid_two.denomination).level == 4 and
                self.is_semi_balanced and
                self.shape[3] > 1 and
                self.stoppers_in_bid_suits and
                self.stoppers_in_unbid_suits() and
                self.nt_level <= 3)

    def _has_doubled_after_responder_passed(self):
        """Return True if double after responder has passed."""
        return (self.hcp <= 20 and
                self.responder_bid_one.is_pass and
                self.bid_two.is_double)

    def _strong_or_responder_bids_minor_at_level_four(self):
        """Return True if strong and responder bids minor at level 4."""
        return ((self.responder_bid_two.level == 4 and
                self.responder_bid_two.denomination.is_minor and
                self.shape[0] >= 6) or
                (self.hcp >= 17 and
                self.responder_bid_two.denomination.is_major and
                (self.five_five or
                 not self.responder_weak_bid())))

    def _has_five_card_major(self):
        """Return True if better than minimum and 5 card major."""
        return (self.hcp >= 14 and
                self.bid_one.denomination.is_major and
                self.shape[0] >= 5 and
                self.next_level(self.bid_one.denomination) <= 4)

    def _responder_bids_three_nt_after_stayman_two_four_card_majors(self):
        """Return True if responder bids 3nt after stayman."""
        return (self.bid_after_stayman and
                self.responder_bid_two.name == '3NT' and
                self.spades >= 4 and
                self.hearts >= 4)

    def _better_than_minimum_and_five_card_suit(self):
        """Return True if better than minimum and five card suit."""
        return (self.hcp >= 13 and
                self.shape[0] == 5 and
                self.nt_level <= 3 and
                self.partner_bid_one.name != '1NT')

    def _minimum_after_stayman(self):
        """Return True if minimum after stayman."""
        return (self.bid_one.name == '1NT' and
                self.responder_bid_one.name == '2C' and
                self.hcp <= 13 and
                self.nt_level <= 3)

    def _maximum_after_stayman(self):
        """Return True if maximum after stayman."""
        return (self.bid_one.name == '1NT' and
                self.responder_bid_one.name == '2C' and
                self.hcp >= 14 and
                self.nt_level <= 3)

    def _has_two_four_card_majors_after_hearts(self):
        """Return True if two four card majors after hearts."""
        return (self.spades >= 4 and
                self.hearts >= 4 and
                self.previous_bid.denomination == self.heart_suit and
                self.spade_suit not in self.opponents_suits)

    def _has_two_four_card_majors_after_spades(self):
        """Return True if two four card majors after spades."""
        return (self.spades >= 4 and
                self.hearts >= 4 and
                self.previous_bid.denomination == self.spade_suit and
                self.heart_suit not in self.opponents_suits)

    def _responder_has_jumped_in_major(self):
        """Return True if Responder has jumped in a major."""
        return (self.responder_bid_one.denomination == self.responder_bid_two.denomination and
                self.responder_bid_one.is_suit_call and
                self.is_jump(self.responder_bid_one, self.responder_bid_two) and
                self.responder_bid_one.denomination.is_major and
                self.suit_length(self.responder_bid_one.denomination) >= 2)

    def _is_unbalanced_and_has_two_card_support_for_responders_major(self):
        """Return True if two card support for responders major."""
        return (self.responder_bid_one.denomination.is_major and
                self.suit_length(self.responder_bid_one.denomination) >= 2 and
                self.hcp >= 13 and
                not self.is_balanced and
                (self.responder_bid_one.level == 2 or
                 self.hcp >= 16))

    def _responders_suit_is_minor_and_semi_balanced(self):
        """Return True if responders suit is minor and semi-balanced."""
        return (self.responder_bid_one.denomination.is_minor and
                self.is_semi_balanced and
                self.nt_level <= 3)

    def _strong_with_stoppers(self):
        """Return True if strong hand and stoppers."""
        return (self.hcp >= 18 and
                self.stoppers_in_unbid_suits() and
                self.nt_level <= 3)

    def _responder_repeats_suit_after_two_nt(self):
        """Return True if responder repeats suit after 2nt."""
        return (self.bid_two.name == '2NT' and
                self.responder_bid_one.denomination == self.responder_bid_two.denomination and
                self.stoppers_in_unbid_suits() and
                self.nt_level <= 3)

    def _partner_bid_minor_at_level_four_two_card_support(self):
        """Return True if responder at level 4 in a minor and 2 card support."""
        return (self.responder_bid_one.denomination.is_minor and
                self.responder_bid_two.level == 4 and
                self.suit_length(self.responder_bid_one.denomination) >= 2)

    def _strong_and_responder_supports_four_card_major(self):
        """Return True if responder supports 4 card major."""
        return (self.hcp >= 17 and
                self.bid_one.denomination.is_major and
                self.suit_length(self.bid_one.denomination) == 4 and
                self.responder_bid_two.denomination == self.bid_one.denomination)

    def _two_four_card_majors_after_stayman(self):
        """Return True if two four card majors."""
        return (self.bid_two.is_major and
                self.hearts >= 4 and
                self.spades >= 4)

    def _can_support_responders_four_card_major(self):
        """Return True if can support responders 4 card major."""
        return (self.opener_bid_one.level == 1 and
                self.suit_length(self.responder_bid_two.denomination) >= 4 and
                self.responder_bid_two.denomination.is_major)

    def _responder_changes_suit_at_level_three(self):
        """Return True if responder_changes_suit_at_level_three."""
        return (self.responder_bid_one.denomination != self.responder_bid_two.denomination and
                self.responder_bid_two.level >= 3)

    def _can_bid_strong_major_at_level_three(self):
        """Return True if strong major."""
        return (self.suit_points(self.longest_suit) >= 5 and
                self.next_level(self.longest_suit) <= 3 and
                self.longest_suit.is_major)

    def _responder_has_jumped_in_second_suit(self):
        """Return True if responder jumps in second suit."""
        return (self.responder_bid_two.denomination == self.opener_suit_two and
                self.responder_bid_two.level == self.opener_bid_two.level+2 and
                self.opener_suit_two.is_major and
                self.next_level(self.opener_suit_two) <= 4)

    def _responder_has_jumped_in_openers_major(self):
        """Return True if responder jumps in openers_major."""
        return (self.responder_bid_two.denomination == self.opener_suit_one and
                self.responder_bid_two.level == self.opener_bid_two.level+2 and
                self.opener_suit_one.is_major and
                self.next_level(self.opener_suit_two) <= 4)

    def _responder_has_jumped_in_openers_minor(self):
        """Return True if responder jumps in minor."""
        return (self.responder_bid_two.denomination == self.opener_suit_one and
                self.responder_bid_two.level >= self.opener_bid_two.level+2 and
                not self.opener_bid_two.is_nt and
                not self.opener_bid_two.is_double and
                self.next_level(self.opener_suit_two) <= 5)

    def _intermediate_with_six_four(self):
        """Return True if responder has replied 1NT and 15 points and 6/4."""
        return (self.responder_bid_one.name == '1NT' and
                self.six_four and
                self.hcp >= 15 and
                self.next_level(self.opener_suit_one) <= 3 and
                self.opener_suit_one.is_major)

    def _responder_repeats_suit_and_can_support(self):
        """Return True if responder moderate and repeats suit."""
        return (self.opener_suit_two == self.responder_bid_two.denomination and
                self.responder_bid_one.denomination == self.responder_bid_two.denomination and
                self.responder_bid_two.level == 3 and
                self.hcp >= 14)

    def _medium_balanced_responder_repeats_at_level_three(self):
        """Return True if medium, balanced, and opener bid suit at level 3."""
        return (self.is_balanced and
                self.responder_bid_two.level <= 3 and
                self.hcp >= 16 and
                self.nt_level <= 3)

    def _can_call_3nt(self):
        """Return True if can call 3NT."""
        return (self.hcp >= 16 and
                (self.partner_bid_one.name != '1NT' or
                 self.hcp >= 18) and
                self.nt_level <= 3)

    def _can_raise_responders_invitation(self):
        """2 card support for responders 6 card suit and better than opening hand."""
        responders_suit = self.responder_bid_one.denomination
        return (self.suit_length(responders_suit) == 2 and
                self.responder_bid_two.level == 3 and
                responders_suit.is_major and self.hcp >= 14)

    def _respsonder_invites_game_in_minor(self):
        """Return True if responder invites game."""
        return (self.responder_bid_two.denomination == self.opener_suit_one and
                self.responder_bid_two.level == 4 and
                self.opener_bid_one.is_minor and
                self.next_level(self.opener_suit_one) <= 5)

    def _responder_promises_stopper(self):
        """Return True if responder promises stopper."""
        return (self.opener_bid_one.name == '1NT' and
                self.responder_bid_two.name == '2NT' and
                self.last_bid.is_pass and self.hcp == 14 and
                self.nt_level <= 3)

    def _responder_jumps_in_major(self):
        """Return True if responder jumps in major."""
        return (self.hcp >= 13 and
                self.partner_bid_one.is_suit_call and
                self.partner_bid_one.denomination == self.partner_bid_two.denomination and
                self.is_jump(self.partner_bid_one, self.partner_bid_two))

    def _responder_has_invited_in_six_card_suit(self):
        """Return True if partner invites in six card suit."""
        return (self.shape[0] >= 6 and
                self.hcp >= 14 and
                self.is_jump(self.opener_bid_two, self.responder_bid_two) and
                self.responder_bid_two.denomination == self.longest_suit and
                self.next_level(self.longest_suit) <= self.game_level(self.longest_suit))

    def _responder_has_invited_in_second_suit(self):
        """Return True if partner invites in second suit."""
        return (self.hcp >= 14 and
                self.is_jump(self.opener_bid_two, self.responder_bid_two) and
                self.opener_suit_two == self.responder_bid_two.denomination and
                self.next_level(self.second_suit) <= self.game_level(self.second_suit) and
                self.second_suit not in self.opponents_suits)



""" Bid for Game
    Acol RespondersLaterBid module
"""
import inspect
from .bidding import Bid, Pass
from .blackwood import Blackwood
from .hand import Hand

inspection = inspect.currentframe

class RespondersLaterBid(Hand):
    """BfG RespondersLaterBid class."""
    def __init__(self, hand_cards, board):
        super(RespondersLaterBid, self).__init__(hand_cards, board)
        self.openers_last_bid = Bid(self.bid_history[-2])
        self.openers_penultimate_bid = Bid(self.bid_history[-6])
        self.openers_first_bid = Bid(self.bid_history[-6])
        if len(self.bid_history) >= 10:
            self.openers_first_bid = Bid(self.bid_history[-10])
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        suit = self.openers_last_bid.denomination

        if self.opener_bid_one.name == '2C':
            bid = self._two_club_opening()

        elif self._has_support_for_openers_new_suit_at_level_four():
            bid = self.next_level_bid(self.opener_suit_three, '3809')

        elif (self.opener_bid_one.is_suit_call and
              self.opener_bid_two.name == '2NT' and
              self.opener_bid_three.name == '3NT' and
              self.hcp >= 16
              and self.nt_level <= 4):
            bid = self.nt_bid(4, '3862')

        # Blackwood
        elif self._opener_jumped_and_can_support():
            bid = self.nt_bid(4, '3806')
        elif self.hcp >= 16 and not self.openers_last_bid.is_pass and self.nt_level <= 4:
            bid = self.nt_bid(4, '3810')
        elif self._opener_strong_and_balanced_and_fifteen_points():
            bid = self.nt_bid(6, '3811')

        # NT in auction
        elif self.openers_last_bid.is_nt and 4 <=self.openers_last_bid.level <= 5:
            bid = self._opener_bid_blackwood()
        elif self.previous_bid.is_nt and self.previous_bid.level >= 4:
            bid = self._responder_bid_nt()
        elif (self.opener_bid_one.is_major and self.suit_holding[self.opener_suit_one] >=3 and
                    self.openers_last_bid.name == '3NT'):
            bid = self.next_level_bid(self.opener_suit_one, '3860')
        elif self._opener_bids_three_nt_after_change_of_major():
            bid = self.next_level_bid(self.opener_suit_one, '3802')
        elif self._opener_bids_three_nt_and_hand_has_six_card_major():
            bid = self.suit_bid(4, self.longest_suit, '3803')
        elif self._five_five_majors_after_3nt():
            bid = self.suit_bid(4, self.heart_suit, '3819')
        elif (self.opener_bid_two.is_nt and
                self.openers_last_bid.denomination == self.bid_one.denomination  and
                self.hcp >= 15 and
                self.nt_level <= 4):
            bid = self.nt_bid(4, '3853')
        elif (self.opener_bid_two.is_nt and
                self.openers_last_bid.denomination == self.bid_one.denomination  and
                self.hcp >= 10 and
                self.nt_level <= 4):
            bid = self.bid_to_game(self.bid_one.denomination, '3876')

        # Misc bids
        elif self._opener_rebids_first_suit_two_card_support():
            bid = self.next_level_bid(self.opener_suit_one, '3808')

        elif self._two_card_support_for_openers_six_card_suit_twelve_points():
            bid = self.next_level_bid(self.opener_suit_one, '3812')

        elif self._opener_support_six_card_suit() and not Bid(self.bid_history[-2]).is_game:
            bid = self.next_level_bid(self.longest_suit, '3813')

        elif self._has_support_for_openers_six_card_suit():
            bid = self.next_level_bid(self.opener_suit_one, '3814')

        elif self._opener_repeats_second_suit():
            bid = self._opener_has_rebid_second_suit()

        elif self._opener_has_bid_same_suit_three_times_and_not_at_game():
            bid = self.bid_to_game(self.opener_suit_one, '3820')

        elif (self.openers_last_bid.denomination == self.bid_two.denomination and
                not self.bid_two.is_pass and
                12 <=self.hcp <= 15 and
                self.nt_level <= 3):
            bid = self.suit_bid(4, self.bid_two.denomination, '3844')

        elif self._can_support_openers_second_suit_and_twelve_points():
            bid = self.next_level_bid(suit, '3805')

        elif self._opener_rebids_second_suit_at_three_level():
            bid = self.suit_bid(4, self.bid_two.denomination, '3804')
        elif (self._opener_has_repeated_suit()and
              self.opener_bid_one.is_minor and
              self.opener_bid_three.level == 4 and
              self.suit_holding[self.opener_suit_one] >= 2 and
              self.nt_level <= 5 and
              self.hcp >= 10):
            bid = self.bid_to_game(self.opener_suit_one, '3878')
        elif (self._opener_has_repeated_suit() and
              self.opener_bid_one.is_major and
              self.opener_bid_three.level == 3 and
              self.suit_holding[self.opener_suit_one] >= 2 and
              self.nt_level <= 4 and
              self.hcp >= 11):
            bid = self.bid_to_game(self.opener_suit_one, '3879')
        elif (not self.bid_one.is_pass and
            self.opener_suit_three == self.bid_one.denomination and
              self.hcp >= 13 and
              self.nt_level <= 3):
            bid = self.bid_to_game(self.bid_one.denomination, '3884')

        # 3NT bids
        elif (self._opener_has_repeated_suit() and
              not (self.opener_suit_one.name == 'NT') and
              1 <= self.suit_holding[self.opener_suit_one] <= 2 and
              self.hcp >= 13 and
              self.stoppers_in_unbid_suits() and
              self.nt_level <= 3):
            bid = self.nt_bid(3, '3894')
        elif self._opener_bid_minor_at_level_three_twelve_points():
            bid = self.nt_bid(3, '3807')
        elif self._opener_at_level_three_nine_points_and_stoppers():
            bid = self.nt_bid(3, '3815')
        elif self._support_after_opener_shows_six_four():
            bid = self.next_level_bid(self.opener_suit_one, '3817')
        elif self._opener_has_bid_two_nt_and_five_four():
            bid = self.nt_bid(3, '3818')
        elif self._opener_is_strong_and_has_bid_two_nt():
            bid = self.nt_bid(3, '3821')
        elif (self.opener_bid_three.name == '2NT' and
              self.stoppers_in_unbid_suits() and
              self.hcp >= 12):
            bid = self.nt_bid(3, '3869')
        elif (self.opener_suit_one != self.opener_suit_three and
              self.opener_suit_one.is_suit and
              self.opener_suit_three.is_suit and
              self.suit_holding[self.opener_suit_one] <= 1 and
              self.suit_holding[self.opener_suit_three] <= 2 and
              self.stoppers_in_unbid_suits()):
            bid = self.nt_bid(3, '3881')
        elif (self.opener_suit_one == self.opener_suit_two and
              self.opener_suit_one != self.opener_suit_three and
              not (self.opener_suit_one.name == 'NT' or self.opener_suit_three.name == 'NT') and
              self.suit_holding[self.opener_suit_one] <= 2 and
              self.stoppers_in_unbid_suits() and
              self.hcp >= 14):
            bid = self.nt_bid(3, '3883')
        else:
            bid = self._select_pass_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _select_pass_bid(self):
        """ Bids the culminate in Pass"""
        # Opener bids slam
        if self.openers_last_bid.level >= 6:
            bid = Pass('3841')

        elif self.openers_last_bid.is_game:
            bid = Pass('3816')

        # Oppenents at game
        elif self._opponents_at_game() and self.bid_history[-2] == 'P':
            bid = Pass('3874')
        elif (self._opponents_at_game() and self.hcp <= 10):
             bid = Pass('3887')

        elif (self.opener_suit_one == self.opener_suit_three and
                self.opener_bid_one.level == self.opener_bid_three.level - 1 and
                self.hcp <= 9):
            bid = Pass('3873')

        # Opener has passed
        elif (self.opener_bid_two.is_pass or self.opener_bid_three.is_pass):
            bid = Pass('3875')

        # Opener repeats suit
        elif (self._opener_has_repeated_suit() and
                self.opener_bid_one.is_suit_call and
                self.suit_holding[self.opener_suit_one] <= 1 and
                self.opener_bid_three.level <= 4 and
                self.hcp <= 18):
            bid = Pass('3877')
        elif (self._opener_has_repeated_suit() and
                self.opener_bid_one.is_suit_call and
                self.nt_level == 4 and
                self.hcp <= 10):
            bid = Pass('3888')
        elif (self.opener_suit_one == self.opener_suit_three and
                self.opener_bid_one.is_minor and
                self.opener_bid_three.level == 4 and
                self.hcp <= 10):
            bid = Pass('3889')

        # Opener has id NT on third bid
        elif self.openers_last_bid.name == '2NT' and self.hcp <= 12:
            bid = Pass('3823')

        # Suit preference
        elif (self.opener_bid_one.is_suit_call and self.opener_bid_three.is_suit_call and
              self.opener_bid_two.is_nt and
              (self.suit_holding[self.opener_suit_three] > self.suit_holding[self.opener_suit_one] + 1 or
               self.suit_holding[self.opener_suit_three] >= 4) and
              self.hcp <= 8):
            bid = Pass('3893')

        elif (self.opener_bid_one.is_suit_call and self.opener_bid_three.is_suit_call and
              (self.suit_holding[self.opener_suit_three] > self.suit_holding[self.opener_suit_one] + 1 or
               self.suit_holding[self.opener_suit_three] >= 4) and
              self.hcp <= 10):
            bid = Pass('3886')

        elif (self.openers_last_bid.denomination == self.openers_first_bid.denomination and
              self.nt_level == 3 and
              (self.openers_last_bid.denomination.is_major and self.hcp <= 10 or
               self.openers_last_bid.denomination.is_minor and self.hcp <= 12)):
            bid = Pass('3843')

        elif (self.opener_suit_one == self.opener_suit_three and
              self.suit_holding[self.opener_suit_one] <= 2):
            bid = Pass('3880')

        elif (self.opener_suit_one == self.opener_suit_two and
              self.opener_suit_one != self.opener_suit_three and
              self.next_level(self.opener_suit_one) >=4  and
              self.hcp <= 10):
            bid = Pass('3891')

        elif (self.opener_bid_one.is_suit_call and self.opener_bid_three.is_suit_call and
              not (self.suit_holding[self.opener_suit_three] > self.suit_holding[self.opener_suit_one] + 1 or
               self.suit_holding[self.opener_suit_three] >= 4) and
              self.next_level(self.opener_suit_one) == self.opener_bid_three.level):
            bid = self.next_level_bid(self.opener_suit_one, '3892')

        # Opener has changed suit on second bid
        elif (self.bid_one.denomination != self.bid_two.denomination and
              self.opener_suit_three == self.bid_one.denomination and
              self.hcp <= 12):
            bid = Pass('3871')

        elif (self.opener_suit_one == self.opener_suit_two and
              self.opener_suit_one != self.opener_suit_three and
              self.opener_bid_one.is_suit_call and
              self.suit_holding[self.opener_suit_one] <= 2 and
              self.opener_suit_three.is_suit and
              self.suit_holding[self.opener_suit_three] <= 3 and
              self.hcp <= 10):
            bid = Pass('3882')

        elif (self.five_five_or_better and
              self.next_level(self.second_suit) <= 3 and
              self.second_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.second_suit, '3895')

        # elif (self.opener_suit_three == self.bid_two.denomination and
        #       self.suit_holding[self.bid_two.denomination]<= 4 and
        #       self.hcp <= 8):
        #     bid = Pass('3890')

        # Weak hand
        elif self.hcp < 6:
            bid = Pass('3872')
        elif self.hcp <= 8 and self.bid_history[-4] == 'P' and self.nt_level >= 4:
            bid = Pass('3896')
        elif self.openers_last_bid.denomination == self.responder_bid_one.denomination and self.hcp <= 12:
            bid = Pass('3897')
        elif (self.opener_bid_one.is_suit_call and
              self .opener_bid_three.is_suit_call and
              self.opener_suit_one != self.opener_suit_three and
              self.suit_holding[self.opener_suit_three] > self.suit_holding[self.opener_suit_one] + 1):
            bid = Pass('3898')
        elif (self.shape[0] >=6 and
              self.suit_points(self.longest_suit) >= 4 and
              self.next_level(self.longest_suit) <= 3):
            bid = self.next_level_bid(self.longest_suit, '3899')
        else:
            bid = Pass('3867')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _opener_bid_blackwood(self):
        """Return bid after opener has bid Blackwood."""
        if self.openers_last_bid.name == '4NT':
            bid = Blackwood(self.cards, self.board).ace_count_bid
        elif self.openers_last_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).king_count_bid
        elif self.openers_last_bid.level >= 6:
            bid = Pass('3848')
        elif self.openers_last_bid.level == 5:
            bid = Pass('3838')
        else:
            bid = Pass('3861')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _responder_bid_nt(self):
        """Return bid after responder has bid NT."""
        if self.previous_bid.name == '4NT':
            bid = self._after_my_4nt()
        elif self.previous_bid.name == '4NT':
            bid = Blackwood(self.cards, self.board).select_contract()
        elif self.previous_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).select_contract()
        elif self.openers_last_bid.level >= 6:
            bid = Pass('3859')
        else:
            bid = Pass('3825')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def suit_preference(self):
        """Return suit preference."""
        suit_one_holding = self.suit_length(self.opener_suit_one)
        suit_two_holding = self.suit_length(self.opener_suit_two)
        if suit_one_holding >= suit_two_holding:
            suit = self.opener_suit_one
        else:
            suit = self.opener_suit_two
        return suit

    def _after_my_4nt(self):
        """Return bid after partner has responded to my 4NT bid."""
        aces = Blackwood(self.cards, self.board).partnership_aces()
        # agreed_suit = None
        if aces == 4:
            bid = self.nt_bid(5, '3826')
        elif (self.opener_suit_one == self.opener_suit_two and
                self.opener_suit_one != self.responder_bid_one.denomination and
                self.suit_length(self.opener_suit_one) >= 2):
            bid = self.suit_bid(6, self.opener_suit_one, '3827')
        elif (aces == 3 and self.hcp < 19 and
              self.openers_last_bid.denomination == self.opener_suit_one):
            bid = Pass('3847')
        elif self.nt_level == 5:
            agreed_suit = Blackwood(self.cards, self.board).agreed_suit
            if not agreed_suit or agreed_suit.name == '':
                print('Suit failure responderts later bid', self.board.__dict__)
            if aces == 3:
                bid = self.next_level_bid(agreed_suit, '3839')
            elif agreed_suit.name:
                bid = self.next_level_bid(agreed_suit, '3854')
            else:
                bid = Pass('0000')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _two_club_opening(self):
        if self._suit_support_after_two_club_opening_opening():
            bid = self.next_level_bid(self.longest_suit, '3829')
        elif self._opener_changes_suit_after_two_club_opening_opening():
            suit = self.suit_preference()
            if self.suit_length(suit) <= 2 and self.nt_level <= 3:
                bid = self.nt_bid(3, '3830')
            else:
                bid = self.next_level_bid(suit, '3831')
        elif self.opener_bid_three.is_nt and self.opener_bid_three.level >= 4:
            bid = self._opener_bid_blackwood()
        elif self.shape[1] >= 5 and self.next_level(self.second_suit) <= 5:
            bid = self.next_level_bid(self.second_suit, '3832')
        elif self.openers_last_bid.level == 6:
            bid = Pass('3851')
        elif self.openers_last_bid.is_game:
            bid = Pass('3850')
        elif (self.suit_holding[self.openers_last_bid.denomination] >= 2 and
                self.next_level(self.openers_last_bid.denomination) <= 4):
            bid = self.next_level_bid(self.openers_last_bid.denomination, '3856')
        elif self.suit_holding[self.openers_last_bid.denomination] <= 1 and self.nt_level <= 3:
            bid = self.next_nt_bid('3855')
        elif (self.suit_holding[self.openers_last_bid.denomination] <= 3 and
                self.openers_last_bid.is_minor and
                self.hcp <= 3):
            bid = Pass('3857')
        elif (self.suit_holding[self.openers_last_bid.denomination] <= 3 and
                self.openers_last_bid.is_minor and
                self.openers_last_bid.level == 4):
            bid = self.next_level_bid(self.openers_last_bid.denomination, '3858')
        else:
            bid = Pass('3833') # This shouldn't happen!!!
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _opener_has_rebid_second_suit(self):
        """Return bid if opener has rebid second suit."""
        if self.hcp >= 15:
            bid = self.nt_bid(4, '3834')
        elif self._can_support_partner_ditributional_hand():
            bid = self.next_level_bid(self.partner_bid_one.denomination, '3835')
        elif self.opener_bid_three.is_game:
            bid = Pass('3836')
        elif self.shape[0] >= 6 and self.longest_suit.is_major:
            bid = self.next_level_bid(self.longest_suit, '3837')
        elif self.opener_suit_three == self.second_suit:
            bid = self.next_level_bid(self.second_suit, '3865')
        else:
            if (self.shape[0] >= 6 and
                    self.suit_points(self.longest_suit) >= 8 and
                    self.next_level(self.longest_suit) <= 4 and
                    self.longest_suit not in self.opponents_suits):
                suit = self.longest_suit
            elif self.opener_suit_two == self.responder_bid_two.denomination:
                suit = self.opener_suit_two
            elif self.suit_length(self.opener_suit_one) >= 2:
                suit = self.opener_suit_one
            elif self.suit_length(self.opener_suit_two) >= 2:
                suit = self.opener_suit_two
            else:
                suit = self.opener_suit_one
            bid = self.next_level_bid(suit, '3864')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _opener_has_repeated_suit(self):
        """Return True id opener bids the same suit three times."""
        return (self.opener_suit_one == self.opener_suit_two and
                self.opener_suit_one == self.opener_suit_three)

    # Various boolean functions

    def _opener_bids_three_nt_after_change_of_major(self):
        """Return True if opener bids 3NT after change of suit."""
        result = (self.opener_bid_three.name == '3NT' and
                  self.opener_suit_one != self.opener_suit_two and
                  self.opener_bid_one.is_suit_call and
                  self.opener_bid_two.is_suit_call and
                  self.opener_suit_one.is_major and
                  self.suit_length(self.opener_suit_one) >= 3)
        return result

    def _opener_bids_three_nt_and_hand_has_six_card_major(self):
        """Return True if opener bids 3NT and hand has 6 card major."""
        result = (self.opener_bid_three.name == '3NT' and
                  self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.longest_suit not in self.opponents_suits and
                  self.next_level(self.longest_suit) <= 4)
        return result

    def _opener_rebids_second_suit_at_three_level(self):
        """Return True if opener rebids second suit at three level."""
        result = (self.opener_suit_three == self.bid_two.denomination and
                  self.bid_one.denomination != self.bid_two.denomination and
                  self.opener_bid_three.is_suit_call and
                  self.opener_bid_three.level == 3 and
                  self.next_level(self.bid_two.denomination) <= 4 and
                  self.hcp >= 9)
        return result

    def _can_support_openers_second_suit_and_twelve_points(self):
        """Return True if can support openers second suit with opening hand."""
        result = (self.openers_last_bid.level == 3 and
                  (self.openers_last_bid.denomination == self.longest_suit or
                   self.openers_last_bid.denomination == self.second_suit) and
                  self.hcp >= 12)
        return result

    def _opener_jumped_and_can_support(self):
        """Return True if opener jumps and can support."""
        result = (self.is_jump(self.bid_one, self.opener_bid_two) and
                  self.bid_one.denomination == self.opener_suit_three and
                  self.hcp >= 16 and
                  not self.openers_last_bid.is_pass and
                  self.nt_level <= 4)
        return result

    def _opener_bid_minor_at_level_three_twelve_points(self):
        """Return True if opener bids minor at level 3."""
        result = (self.opener_bid_three.level == 3 and
                  self.opener_suit_three.is_minor and
                  self.hcp >= 12 and
                  self.is_balanced and
                  self.nt_level <= 3)
        return result

    def _opener_rebids_first_suit_two_card_support(self):
        """Return True if opener rebids opening suit and two card support."""
        result = (self.opener_suit_one == self.opener_suit_three and
                  self.opener_suit_one != self.opener_suit_two and
                  not self.responder_bid_two.is_nt and
                  self.suit_length(self.opener_suit_one) >= 2 and
                  self.opener_bid_three.level <= 4 and
                  not self.opener_bid_three.is_game and
                  self.hcp >= 8)
        return result

    def _suit_support_after_two_club_opening_opening(self):
        """Return True if after 2C can support."""
        result = (self.opener_bid_one.name == '2C' and
                  self.hcp >= 8 and
                  (self.opener_suit_two == self.longest_suit or
                   self.opener_suit_three == self.longest_suit) and
                  self.shape[0] >= 6)
        return result

    def _opener_changes_suit_after_two_club_opening_opening(self):
        """Return True if opener changes suit after 2C opening."""
        result = (self.opener_bid_one.name == '2C' and
                  self.opener_suit_two != self.opener_suit_three and
                  not self.openers_last_bid.is_game)
        return result

    def _opener_strong_and_balanced_and_fifteen_points(self):
        """Return True if opener is strong and balanced."""
        result = (self.opener_bid_two.name == '2NT' and
                  self.opener_bid_three.name == '3NT' and
                  self.hcp >= 15)
        return result

    def _two_card_support_for_openers_six_card_suit_twelve_points(self):
        """Return True if two card support for opener's six card suit."""
        result = (self.opener_suit_one.is_major and
                  self._opener_has_repeated_suit() and
                  self.suit_length(self.opener_suit_one) >= 2 and
                  self.hcp >= 12 and
                  self.next_level(self.opener_suit_one) <= 4)
        return result

    def _opener_at_level_three_nine_points_and_stoppers(self):
        """Return True if nine points and stoppers."""
        result = (self.openers_last_bid.level == 3 and
                  not self._opener_has_repeated_suit() and
                  self.hcp >= 11 and
                  self.stoppers_in_unbid_suits() and
                  self.nt_level <= 3)
        return result

    def _opener_repeats_second_suit(self):
        """Return True if opener repeats second_suit."""
        result = (self.opener_bid_two.is_suit_call and
                  self.opener_suit_one != self.opener_suit_two and
                  self.opener_suit_two == self.opener_suit_three
                  and self.nt_level <= 4)
        return result

    def _opener_support_six_card_suit(self):
        """Return True if opener support six card suit."""
        result = (self.opener_suit_three == self.longest_suit and
                  self.shape[0] >= 6 and
                  self.hcp >= 8 and
                  self.next_level(self.longest_suit) <= 6)
        return result

    def _has_support_for_openers_new_suit_at_level_four(self):
        """Return True if opener shows new suit at 4 level and support."""
        result = (self.opener_bid_three.level == 4 and
                  self.opener_suit_three != self.opener_suit_one and
                  self.opener_suit_three != self.opener_suit_two and
                  self.suit_length(self.opener_suit_three) >= 3 and
                  not self.opener_bid_three.is_game)
        return result

    def _support_after_opener_shows_six_four(self):
        """Return True if support for opener who shows six four."""
        result = (self.hcp >= 9 and
                  self.opener_suit_one == self.opener_suit_two and
                  self.opener_suit_one != self.opener_suit_three and
                  not self.opener_bid_three.is_nt and
                  # self.responder_bid_two.is_nt and
                  not Bid(self.bid_history[-6]).is_pass and
                  self.suit_length(self.opener_suit_one) >= 2 and
                  not (self.openers_last_bid.denomination == self.opener_suit_one and
                       self.openers_last_bid.is_game))
        return result

    def _opener_has_bid_two_nt_and_five_four(self):
        """Return True if opener hand bid 2NT and hand is 5/4."""
        result = (self.opener_bid_three.name == '2NT' and
                  self.five_four and
                  self.hcp >= 10 and
                  self.nt_level <= 3)
        return result

    def _has_support_for_openers_six_card_suit(self):
        """Return True if opener has 6 card suit and support."""
        result = (self.opener_suit_one == self.opener_suit_three and
                  self.opener_suit_one != self.opener_suit_two and
                  self.suit_length(self.opener_suit_one) >= 2 and
                  self.hcp >= 12 and
                  self.opener_bid_three.level <= 6 and
                  self.next_level(self.opener_suit_one) <= 5 and
                  not self.opener_bid_three.is_game)
        return result

    def _five_five_majors_after_3nt(self):
        """Return True if _five_five_majors_after_3nt."""
        result = (self.openers_last_bid.name == '3NT' and
                  self.spades >= 5 and self.hearts >= 5 and
                  self.next_level(self.heart_suit) <= 4 and
                  self.heart_suit not in self.opponents_suits)
        return result

    def _opener_has_bid_same_suit_three_times_and_not_at_game(self):
        """Return True if opener has bid suit 3 times and not at game."""
        result = (self._opener_has_repeated_suit() and
                  self.suit_length(self.opener_suit_one) >= 3 and
                  not self.bidding_above_game and
                  self.hcp >= 12)
        return result

    def _opener_is_strong_and_has_bid_two_nt(self):
        """Return True if opener is strong and has bid 2NT."""
        result = (self.opener_bid_three.name == '2NT' and
                  self.hcp >= 8 and
                  self.opener_suit_one != self.opener_suit_two)
        return result

    def _can_support_partner_ditributional_hand(self):
        """Return True partner has shown 5/5 hand."""
        result = (self.partner_bid_one.denomination.is_suit and
                    self.partner_bid_one.denomination != self.partner_bid_two.denomination and
                    self.openers_last_bid.denomination == self.partner_bid_two.denomination and
                    self.suit_length(self.partner_bid_one.denomination) > self.suit_length(self.partner_bid_two.denomination))
        return result



""" Bid for Game Hand module."""
import os
from termcolor import cprint
from pathlib import Path
import json
import inspect

from bridgeobjects import (Board, Hand, Card, Call, Suit, NoTrumps,
                           SUIT_NAMES, CALLS, SUITS, Denomination)
from bfgbidding.bidding import Bid
from bfgbidding.utils import NamedBids

MODULE_COLOUR = 'blue'
TRACE_COLOUR = 'cyan'

MODULE_TRACE = 'batch_tests/data//module_trace.txt'
TRACE_FILE = 'trace.txt'
TRACE_PATH = Path(os.getcwd(), TRACE_FILE)
trace_set = False
trace_value = False


class Hand(Hand):
    """A sub class of bridgeobjects Hand, to enable bidding."""
    overcaller_position = {'none': 0, 'second_seat': 1, 'fourth_seat': 2}

    spade_suit = SUITS['S']
    heart_suit = SUITS['H']
    diamond_suit = SUITS['D']
    club_suit = SUITS['C']

    def __init__(
            self,
            hand_cards: list[Card] = None,
            board: Board = None,
            display_trace: bool = False,
            display_hand: bool = False,
            *args, **kwargs) -> None:
        super().__init__(hand_cards, *args, **kwargs)
        (self.board, bid_history, self.overcaller) = self._setup_board(board)
        self.bid_history = bid_history
        self.suits = [SUITS[name] for name in 'CDHS']
        self.no_trumps = NoTrumps()
        self._losers = None
        self.display_trace = display_trace
        self.display_hand = display_hand

        self.holding_partner_one = 0
        self.unplayed_cards = [card for card in self.cards]

        utils = NamedBids(self, self.bid_history)
        self.opener_bid_one = utils.opener_bid_one
        self.opener_bid_two = utils.opener_bid_two
        self.opener_bid_three = utils.opener_bid_three
        self.responder_bid_one = utils.responder_bid_one
        self.responder_bid_two = utils.responder_bid_two
        self.responder_bid_three = utils.responder_bid_three
        self.overcaller_bid_one = utils.overcaller_bid_one
        self.overcaller_bid_two = utils.overcaller_bid_two
        self.overcallers_last_bid = utils.overcallers_last_bid
        self.advancer_bid_one = utils.advancer_bid_one
        self.advancer_bid_two = utils.advancer_bid_two
        self.previous_bid = utils.previous_bid
        self.last_bid = utils.last_bid
        self.right_hand_bid = utils.right_hand_bid
        self.my_last_bid = utils.my_last_bid
        self.partner_bid_one = utils.partner_bid_one
        self.partner_bid_two = utils.partner_bid_two
        self.partner_penultimate_bid = utils.partner_penultimate_bid
        self.partner_last_bid = utils.partner_last_bid
        self.bid_one = utils.bid_one
        self.bid_two = utils.bid_two
        self.bid_three = utils.bid_three

        self.responders_support = utils.responders_support
        self.overcaller_has_jumped = utils.overcaller_has_jumped
        self.bid_after_stayman = utils.bid_after_stayman
        self.overcaller_in_second_seat = utils.overcaller_in_second_seat
        self.overcaller_in_fourth_seat = utils.overcaller_in_fourth_seat

        self.opener_suit_one = self.opener_bid_one.denomination
        self.opener_suit_two = self.opener_bid_two.denomination
        self.opener_suit_three = self.opener_bid_three.denomination
        self.suit_one = utils.suit_one
        self.overcaller_suit_one = utils.overcaller_suit_one
        self.advancer_suit_one = utils.advancer_suit_one
        self.advancer_suit_two = utils.advancer_suit_two
        self.holding_partner_one = utils.holding_partner_one

        (self.trace_module, self.last_hand) = self._get_trace_module()

    @staticmethod
    def _setup_board(board: Board | None) -> tuple[Board, list[Call], bool]:
        """Initialise the board."""
        if not board:
            board = Board()
            board.active_bid_history = []
        bid_history = []
        overcaller = False
        bid_history = board.active_bid_history
        if len(bid_history) % 2 != 0:
            overcaller = True
        return (board, bid_history, overcaller)

    def to_json(self) -> str:
        """Return object as json string property."""
        json_str = json.dumps({
            'cards': [card.name for card in self.cards],
            'unplayed_cards': [card.name for card in self.unplayed_cards],
        })
        return json_str

    def from_json(self, json_str: str) -> None:
        """Populate the attributes from the json string."""
        hand_dict = json.loads(json_str)
        self.cards = [Card(name) for name in hand_dict['cards']]
        self.unplayed_cards = [Card(name)
                               for name in hand_dict['unplayed_cards']]

    @property
    def nt_level(self) -> int:
        """Return the level of no trumps"""
        return self.next_nt_bid().level

    @property
    def losers(self) -> int:
        """Return the number of losers in the Hand."""
        if not self._losers:
            self._losers = self._get_losers()
        return self._losers

    def _get_losers(self) -> int:
        """Calculate and return the number of losers in the Hand."""
        losers = 0
        for suit in SUIT_NAMES:
            cards = [card for card in self.cards_by_suit[suit]]
            honour_count = 0
            for card in cards:
                if card.value >= 12:
                    honour_count += 1
                elif card.value == 11:
                    if self.suit_points(suit) > 2:
                        honour_count += 1
            count = min(3, len(cards))
            losers += count - honour_count
        return losers

    def is_insufficient_bid(self, test_bid) -> bool:
        """Return True if the bid is insufficient."""
        last_bid = None
        for last_bid in self.bid_history[::-1]:
            if Bid(last_bid).is_value_call:
                break
        last_bid_index = CALLS.index(last_bid)
        test_bid_index = CALLS.index(test_bid.name)
        if test_bid_index <= last_bid_index:
            return True
        return False

    @property
    def singleton_honour(self) -> bool:
        """Return True if hand has singleton K or Q."""
        return self._singleton_honour()

    def _singleton_honour(self) -> bool:
        """Return True if hand has singleton K or Q."""
        if self.shape[3] == 1:
            for suit_name in SUIT_NAMES:
                suit = Suit(suit_name)
                if self.suit_length(suit) == 1:
                    for card in self.cards:
                        if card.suit == suit:
                            if card.rank in 'KQ':
                                return True
        return False

    @property
    def opponents_have_bid(self) -> bool:
        """Return True if the opponents have made a bid."""
        return self._opponents_have_bid()

    def _opponents_have_bid(self) -> bool:
        """Return True if the opponents have made a bid."""
        for bid in self.bid_history[1::2]:
            if Bid(bid).is_value_call:
                return True
        return False

    @property
    def competitive_auction(self) -> bool:
        """Return True if the last bid is not Pass or Double."""
        return self._competitive_auction()

    def _competitive_auction(self):
        """Return True if the last bid is not Pass or Double."""
        value = False
        if len(self.bid_history) >= 1:
            value = Bid(self.bid_history[-1]).is_value_call
        if len(self.bid_history) >= 3:
            value = value or Bid(self.bid_history[-3]).is_value_call
        if len(self.bid_history) >= 5:
            value = value or Bid(self.bid_history[-5]).is_value_call
        return value

    @staticmethod
    def _is_value_call_or_double(bid: Bid) -> bool:
        """Return True if is value call or double."""
        call = Bid(bid)
        value = call.is_value_call or call.is_double
        return value

    @property
    def opponents_have_doubled(self) -> bool:
        """Return True if opponents have doubled with
        no further bid from them."""
        return self._opponents_have_doubled()

    def _opponents_have_doubled(self) -> bool:
        """Return True if opponents have doubled with
        no further bid from them."""
        if len(self.bid_history) >= 1:
            if Bid(self.bid_history[-1]).is_double:
                return True
            elif len(self.bid_history) >= 3:
                if (Bid(self.bid_history[-3]).is_double and
                        Bid(self.bid_history[-1]).is_pass):
                    return True
        return False

    @property
    def partner_has_passed(self) -> bool:
        """Return True of partner passed on first round."""
        return self._partner_has_passed()

    def _partner_has_passed(self) -> bool:
        """Return True of partner passed on first round."""
        value = False
        if len(self.bid_history) >= 5:
            if Bid(self.bid_history[-5]).is_pass:
                value = True
        return value

    def _bid_after_stayman(self) -> bool:
        """Return True if responder has bid Clubs after NT opening."""
        if (self.opener_bid_one.is_nt and
                self.responder_bid_one.denomination == self.club_suit):
            return True
        return False

# TODO: sort out use of shortage points
    def suit_bid(self, level: int,
                 suit: Suit,
                 comment: str = '0000',
                 use_shortage_points: bool = False) -> Bid:
        """Return a bid in the suit at the given level."""
        return Bid(
            self._call_name(level, suit.name), comment, use_shortage_points)

    def heart_bid(self,
                  level: int,
                  comment: str = '0000',
                  use_shortage_points: bool = False) -> Bid:
        """Return a bid in Hearts at the given level."""
        return Bid(self._call_name(level, 'H'), comment, use_shortage_points)

    def spade_bid(self,
                  level: int,
                  comment: str = '0000',
                  use_shortage_points: bool = False) -> Bid:
        """Return a bid in Spades at the given level."""
        return Bid(self._call_name(level, 'S'), comment, use_shortage_points)

    def club_bid(self,
                 level: int,
                 comment: str = '0000',
                 use_shortage_points: bool = False) -> Bid:
        """Return a bid in Clubs at the given level."""
        return Bid(self._call_name(level, 'C'), comment, use_shortage_points)

    def diamond_bid(self,
                    level: int,
                    comment: str = '0000',
                    use_shortage_points: bool = False) -> Bid:
        """Return a bid in Diamonds at the given level."""
        return Bid(self._call_name(level, 'D'), comment, use_shortage_points)

    def nt_bid(self,
               level: int,
               comment: str = '0000',
               use_shortage_points: bool = False) -> Bid:
        """Return a bid in NT at the given level."""
        return Bid(self._call_name(level, 'NT'), comment, use_shortage_points)

    def barrier_is_broken(self, first_bid: Bid, second_bid: Bid) -> bool:
        """Return True if second_bid breaks the barrier relative to bid_one."""
        level_one = first_bid.level
        barrier = Bid(
            self._call_name(level_one+1, first_bid.denomination.name))
        return self._higher_bid(barrier, second_bid)

    def cheapest_long_suit(self) -> Suit | None:
        """Return the longest suit or cheapest of equal length suits to bid."""
        if (self.shape[0] > self.shape[1]
                and self.longest_suit not in self.opponents_suits):
            return self.longest_suit

        max_length = self.shape[1]
        last_bid_denomination = self._last_denomination_called()
        extended_suit_list = self._get_extended_suit_list()

        # find the starting index in the extended suit list
        index = extended_suit_list.index(last_bid_denomination)

        # find the level of NT in the middle of the extended suit list
        nt_level = self.nt_level
        while index < len(extended_suit_list) - 1 and index < 8:
            index += 1
            if index >= 4:
                # if we're in the second half, increment the level by 1.
                # I.e. we have past NT and onto the next level
                level = nt_level + 1
            else:
                level = nt_level

            # test the next suit
            suit = extended_suit_list[index]
            if (self.suit_length(suit) == max_length and
                    self.next_level(suit) == level and
                    suit not in self.opponents_suits):
                return suit
        return None

    def _last_denomination_called(self) -> Denomination:
        """Return the denomination of the last value call."""
        for bid in self.bid_history[::-1]:
            call = Bid(bid)
            if call.is_suit_call:
                return call.denomination
        assert False, 'Last denomination called with no value call in history'

    @staticmethod
    def _get_extended_suit_list() -> list[str]:
        """Return an extended list of sorted of suits,
            e.g. [C, D, H, S, C, D, H, S]."""
        extended_suit_list = [Suit(name) for name in SUIT_NAMES]
        extended_suit_list.sort(key=lambda x: x.rank)

        # get a double suit list, e.g. [C, D, H, S, C, D, H, S]
        extended_suit_list.extend(extended_suit_list)
        return extended_suit_list

    @staticmethod
    def _higher_bid(first_bid: Bid, second_bid: Bid) -> bool:
        """Return True if first_bid is the lower."""
        level_one = first_bid.level
        level_two = second_bid.level
        if level_two > level_one:
            return True
        else:
            if level_one == level_two:
                if first_bid.denomination < second_bid.denomination:
                    return True
        return False

    @staticmethod
    def is_jump(first_bid: Bid, second_bid: Bid) -> bool:
        """Return True if second bid is a jump over first."""
        # e.g.  first_bid = Bid('1S')
        #       second_bid = Bid('2NT')
        if second_bid.is_value_call and first_bid.is_value_call:
            jump_level = second_bid.level - first_bid.level
            if jump_level > 1:
                return True
            elif jump_level == 1:
                if second_bid.is_suit_call and first_bid.is_suit_call:
                    if second_bid.denomination > first_bid.denomination:
                        return True
                elif second_bid.is_nt:
                    return True
        return False

    def jump_bid_made(self, test_bid: Bid) -> bool:
        """Test bid against bid-history and return True
                    if test_bid is a jump over last relevant bid."""
        second_bid = None
        for bid in self.bid_history[::-1]:
            if bid != 'P' and bid != 'D' and bid != 'R':
                second_bid = Bid(bid)
                break
        jump_level = test_bid.level - second_bid.level
        if jump_level > 1:
            return True
        elif (jump_level == 1
                and second_bid.denomination < test_bid.denomination):
            return True
        return False

    def levels_to_bid(self, bid_one: Bid, bid_two: Bid) -> int:
        """
        Return the number of 'steps' between bid_one and bid_two.

        e.g.

        1C to 3H is 3 steps: (1C, 2C 3C, 3H) but
        1S to 3H is 2 steps: (1S, 2S, 3H).
        """
        steps = 0
        while bid_one < bid_two:
            bid_one = self.suit_bid(bid_one.level+1, bid_one.denomination)
            steps += 1
        return steps

    @property
    def overcall_made(self) -> bool:
        """Return 1 if an overcall has been made by 2nd and not by 4th seat.
        Return 2 if overcall has been made in 4th seat.
        """
        return self._overcall_made()

    def _overcall_made(self) -> bool:
        """Return 1 if an overcall has been made by 2nd and not by 4th seat.
        Return 2 if overcall has been made in 4th seat.
        """
        bid_history = self.bid_history
        for bid in bid_history[1::2]:
            if not Bid(bid).is_pass:
                if Bid(bid_history[-1]).is_value_call and len(bid_history) > 2:
                    return self.overcaller_position['fourth_seat']
                return self.overcaller_position['second_seat']
        return self.overcaller_position['none']

    @property
    def opponents_at_game(self) -> bool:
        """Return True if opponents at game level."""
        return self._opponents_at_game()

    def _opponents_at_game(self):
        """Return True if opponents at game level."""
        if len(self.bid_history) >= 3:
            if (Bid(self.bid_history[-3]).is_game or
                    Bid(self.bid_history[-1]).is_game):
                return True
        elif len(self.bid_history) >= 1:
            if Bid(self.bid_history[-1]).is_game:
                return True
        return False

    @property
    def bidding_above_game(self) -> bool:
        """Return True if the bidding is at or above game level."""
        return self._bidding_above_game()

    def _bidding_above_game(self):
        """Return True if the bidding is at or above game level."""
        for bid in self.bid_history[::1]:
            if Bid(bid).is_game:
                return True
        return False

    @property
    def can_double(self) -> bool:
        """Return True if double is legal."""
        return self._can_double()

    def _can_double(self) -> bool:
        if self.bid_history[-1] != 'P' or self.bid_history[-1] != 'P':
            return True
        return False

    @property
    def partner_doubled_game(self) -> bool:
        """Return True if partner has doubled at or above game level."""
        return self._partner_doubled_game()

    def _partner_doubled_game(self) -> bool:
        """Return True if partner has doubled at or above game level."""
        if 'D' in self.bid_history:
            index = self.bid_history.index('D')
            bids = [Bid(bid) for bid in self.bid_history[index::-1]]
            for bid in bids:
                if bid.is_value_call:
                    return bid.is_game
        return False

    def double_level(self) -> int:
        """Return the level at which the DOUBLE was made."""
        level = 0
        index = self.bid_history.index('D')
        bids = [Bid(bid) for bid in self.bid_history[index::-1]]
        for bid in bids:
            if bid.is_value_call:
                level = bid.level
                break
        return level

    @property
    def opponents_suits(self) -> list[Suit]:
        """Return a list of all opponent's bid suits."""
        return self._opponents_suits()

    def _opponents_suits(self) -> list[Suit]:
        """Return a list of all opponent's bid suits."""
        opponents_suits = []
        start = 1
        if self.overcaller:
            start = 0
        for bid in self.bid_history[start::2]:
            call = Bid(bid)
            if call.is_suit_call:
                opponents_suits.append(Suit(call.denomination.name))
        return opponents_suits

    def partner_bids_at_lowest_level(self) -> bool:
        """Return True if partner has not jumped."""
        double_index = None
        for index, bid in enumerate(self.bid_history):
            if Bid(bid).is_double:
                double_index = index
                break

        value = False
        partners_bid = Bid(self.bid_history[double_index+2])
        for bid in list(reversed(self.bid_history))[double_index+1::2]:
            if Bid(bid).is_value_call:
                # noinspection PyTypeChecker
                value = self.is_jump(Bid(bid), partners_bid)
        return value

    def openers_agreed_suit(self) -> Suit:
        """Return a suit agreed by opener."""
        suit = None
        if len(self.bid_history) >= 5:
            opps_bid_one = Bid(self.bid_history[0], '')
            opps_bid_two = Bid(self.bid_history[2], '')
            opps_bid_three = Bid(self.bid_history[4], '')
            if ((opps_bid_two.denomination == opps_bid_three.denomination) or
                    (opps_bid_one.denomination == opps_bid_two.denomination)):
                suit = opps_bid_two.denomination
            if opps_bid_one.is_double and opps_bid_two.is_suit_call:
                suit = opps_bid_two.denomination
        return suit

    def cheaper_suit(self, suit_one: Suit, suit_two: Suit) -> Suit:
        """Return the cheaper of two suits based on bid history."""
        suits = [suit_one, suit_two]
        suit = self.cheapest_suit(suits)
        return suit

    def next_four_card_suit(self) -> Suit:
        """Return the cheapest four card suit."""
        suits = []
        for suit in self.suits_by_length:
            if self.suit_length(suit) == 4:
                suits.append(suit)
        suit = self.cheapest_suit(suits)
        return suit

    def cheapest_suit(self, suits: list[Suit]) -> Suit:
        """Return the cheapest suit based on bid history."""
        suit_names = [(suit.name, suit.rank) for suit in suits]
        sorted_suits = sorted(suit_names, key=lambda tup: tup[1])
        suits = [Suit(suit[0]) for suit in sorted_suits]
        last_rank = self._last_bid_rank()
        if not suits:
            return None
        for suit in suits:
            if suit.rank > last_rank:
                break
        else:
            suit = suits[0]
        return suit

    def _last_bid_rank(self) -> int:
        """Return the rank of the last value bid."""
        bid = None
        for bid in self.bid_history[::-1]:
            if Call(bid).is_value_call:
                break
        last_rank = Call(bid).denomination.rank
        if last_rank == 4:
            last_rank = -1
        return last_rank

    @staticmethod
    def game_level(suit: Suit) -> int:
        """Return the level of game in the suit."""
        if suit.is_major:
            level = 4
        elif suit.is_minor:
            level = 5
        else:
            level = 3
        return level

    def bid_to_game(self,
                    denomination: Denomination,
                    comment: str = '0000',
                    use_distribution_points: bool = False) -> Bid | None:
        """Return game level bid in given denomination"""
        if denomination.is_nt:
            return Bid('3NT', comment, use_distribution_points)
        elif denomination.is_major:
            return Bid(
                    self._call_name(4,  denomination.name),
                    comment, use_distribution_points)
        elif denomination.is_minor:
            return Bid(
                self._call_name(5,  denomination.name),
                comment, use_distribution_points)
        return None

    @property
    def stoppers_in_bid_suits(self) -> bool:
        """Return True if hand contains stoppers
            in all opponent's bid suits."""
        return self._stoppers_in_bid_suits()

    @property
    def poor_stoppers_in_bid_suits(self) -> bool:
        """Return True if hand contains stoppers (including ten)
            in all opponent's bid suits."""
        return self._stoppers_in_bid_suits(lowest_card='T')

    def _stoppers_in_bid_suits(self, lowest_card='J'):
        """Return True if hand contains stoppers
            in all opponent's bid suits."""
        for suit in self.opponents_suits:
            if not self.suit_stopper(suit, lowest_card):
                return False
        return True

    def suit_stopper(self, suit: Suit, lowest_card: str = 'J') -> bool:
        """Return True if the hand contains a stopper in 'suit'."""
        if not suit:
            return False
        if not suit.is_suit:
            return False

        suit_holding = self.suit_holding
        if suit_holding[suit] >= 5:
            return True

        poor_stopper = False
        ace_stopper = Card('A', suit.name) in self.cards
        king_stopper = (Card('K', suit.name) in self.cards
                        and suit_holding[suit] >= 2)
        queen_stopper = (Card('Q', suit.name) in self.cards
                         and suit_holding[suit] >= 3)
        jack_stopper = (Card('J', suit.name) in self.cards
                        and suit_holding[suit] >= 4)
        if lowest_card == 'T':
            poor_stopper = (Card('T', suit.name) in self.cards
                            and suit_holding[suit] >= 4)
        if (ace_stopper
                or king_stopper
                or queen_stopper
                or jack_stopper
                or poor_stopper):
            return True
        return False

    @property
    def unbid_suit(self) -> Suit:
        """Return the unbid _suit (if any) or None."""
        return self._unbid_suit()

    def _unbid_suit(self) -> Suit:
        """Return the unbid _suit (if any) or None."""
        suits = [suit for suit in self.suits]
        for bid in self.bid_history:
            bid_suit = Bid(bid).denomination
            if bid_suit in suits:
                suits.remove(bid_suit)
        if len(suits) == 1:
            return suits[0]
        return None

    def stoppers_in_unbid_suits(self) -> bool:
        """Return True if hand contains stoppers in all
            opponent's unbid suits."""
        bid_suits = []
        for bid in self.bid_history:
            bid_suit = Bid(bid).denomination
            if bid_suit.is_suit:
                bid_suits.append(bid_suit)
        for suit in self.suits:
            if suit not in bid_suits:
                if not self.suit_stopper(suit):
                    return False
        return True

    def stoppers_in_other_suits(self, suit) -> bool:
        """Return True if hand has stoppers in all suits except suit."""
        for test_suit in self.suits:
            if test_suit != suit:
                if not self.suit_stopper(test_suit):
                    return False
        return True

    def four_in_bid_suits(self, lowest_card='J') -> bool:
        """Return True if hand contains stoppers or 4 cards in all
            opponent's bid suits.
        """
        for suit in self.opponents_suits:
            if (not self.suit_stopper(suit, lowest_card) and
                    self.suit_holding[suit] <= 3):
                return False
        return True

    def three_suits_bid_and_stopper(self) -> bool:
        """Returns True if three suits bid and
                hand has an stopper in the unbid suit"""
        suits_bid = [False, False, False, False]
        for bid_name in self.bid_history[::2]:
            bid = Bid(bid_name)
            if bid.is_suit_call:
                suits_bid[bid.denomination.rank] = True
                suits_bid[bid.denomination.rank] = True
        if suits_bid.count(True) == 3:
            for index, suit_bid in enumerate(suits_bid):
                if not suit_bid:
                    suit = self.suits[index]
                    if self.suit_stopper(suit):
                        return True
                    if (self.suit_points(suit) >= 1
                            and self.suit_holding[suit] >= 3):
                        return True
        return False

    def can_bid_suit(self, suit) -> bool:
        """Return False if suit is in opponents bids."""
        for bid in self.bid_history[::-1][::2]:
            if Bid(bid).is_suit_call:
                if suit == Bid(bid).denomination:
                    return False
        return True

    def has_stopper(self, suit) -> bool:
        """Return self.suit_stopper."""
        return self.suit_stopper(suit)

    def next_level(self, suit: Suit, raise_level: int = 0) -> bool:
        """Return the next next level for a suit bid."""
        level = self.next_level_bid(suit, '000', raise_level).level
        return level

    def current_bid_level(self) -> int:
        """Return the level of the latest quantitative bid."""
        bid = self._get_last_bid()
        return bid.level

    def next_level_bid(self,
                       suit: Suit,
                       comment: str = '0000',
                       raise_level: int = 0) -> int:
        """Return the lowest possible bid in suit."""
        last_bid = self._get_last_bid()
        level = self._get_Level_of_last_bid(last_bid)
        level += raise_level
        if last_bid.is_nt:
            level += 1
        elif suit and suit.is_suit:
            if last_bid.denomination >= suit:
                level += 1
        if level > 7:
            level = 7
        new_bid = Bid(self._call_name(level, suit.name), comment)
        return new_bid

    def _get_Level_of_last_bid(self, last_bid: Bid) -> int:
        """Return the level of the last bid."""
        if last_bid is None:
            return 1
        return last_bid.level

    def _get_last_bid(self) -> Bid | None:
        """Return last quantitative bid from history."""
        for bid_level in self.bid_history[::-1]:
            if bid_level != 'P' and bid_level != 'D' and bid_level != 'R':
                return Bid(bid_level, '0000')
        return None

    def next_nt_bid(self, comment: str = '0000', raise_level: int = 0) -> Bid:
        """Return the lowest possible bid in no trumps."""
        denomination = self.no_trumps
        bid = self.next_level_bid(denomination, comment, raise_level)
        return bid

    def responder_weak_bid(self) -> bool:
        """Responder has shown preference at  the lowest level."""
        overcallers_last_bid = Call(self.bid_history[-3])
        return (overcallers_last_bid.is_pass and
                self.partner_bid_one.is_pass and
                (self.partner_last_bid.denomination == self.opener_suit_one or
                 self.partner_last_bid.denomination == self.opener_suit_two)
                and not self.is_jump(
                    self.opener_bid_two, self.partner_last_bid))

    def advancer_preference(self, call_id: str = '0000') -> Bid:
        """Respond after a 3 level bid make suit preference."""
        suit_one = self.overcaller_bid_one.denomination
        suit_two = self.overcaller_bid_two.denomination
        if self.overcaller_bid_one.is_double:
            suit_one = self.overcaller_bid_two.denomination
            suit_two = self.overcaller_bid_three.denomination

        if suit_one.is_major and self.suit_length(suit_one) >= 3:
            suit = suit_one
        elif self.suit_holding[suit_one] + 1 >= self.suit_holding[suit_two]:
            suit = suit_one
        else:
            suit = suit_two

        raise_level = 0
        if self.hcp >= 8 and self.next_level(suit) <= 3:
            raise_level = 1
        bid = self.next_level_bid(suit, call_id, raise_level=raise_level)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def unbid_four_card_major(self) -> Suit | None:
        """Return an unbid four card major or None."""
        bid_suits = [Bid(bid).denomination
                     for bid in self.bid_history if Bid(bid).is_suit_call]
        if (self.hearts >= 4 and self.heart_suit not in bid_suits):
            return self.heart_suit
        elif (self.spades >= 4 and self.spade_suit not in bid_suits):
            return self.spade_suit
        return None

    @staticmethod
    def quantitative_raise(
            points: int,
            base_level: int,
            point_list: list[int],
            maximum_level: int = 5) -> int:
        """
            Return a bid based on a quantitative raise.
            max raise is the number of elements in point_list 1, 2,3 or 4
            scan the (reversed) points list until the points in
            the hand exceeds the level
            This shows whether it is a 3,2 or 1 raise etc.

            e.g.
            level = self.quantitative_raise(points, 1, [6, 10, 13, 16], 5)
            if points = 11 this raises level = 1+2 = 3.
        """
        maximum_raise = len(point_list)
        point_list = list(reversed(point_list))
        raise_level = 0
        for index, item in enumerate(point_list):
            if points >= item:
                raise_level = base_level + maximum_raise - index
                break
        if raise_level > maximum_level:
            raise_level = maximum_level
        return raise_level

    def hand_value_points(self, bid_suit: Suit) -> int:
        """Return the hand value points for the given suit."""
        hand_value_points = (self.hcp + self.support_shape_points(bid_suit))
        return hand_value_points

    def support_points(self, bidders_suit: Suit) -> int:
        """
        Calculate the sum of high card points and distribution points
        based on support_shape_points.
        """
        return self.high_card_points + self.support_shape_points(bidders_suit)

    def support_shape_points(self, bidders_suit: Suit) -> int:
        """
        Calculate the distribution points based on
        3 for a void, 2 for a singleton and 1 for a doubleton.
        """
        points = 0
        if bidders_suit.is_suit:
            if self._suit_support(bidders_suit):
                for index in range(4):
                    if self.shape[index] < 3:
                        points += 3 - self.shape[index]
        return points

    def _suit_support(self, bid_suit: Suit) -> bool:
        """Return True if the hand contains at least 3 of bid_suit."""
        if self.suit_holding[bid_suit] >= 3:
            return True
        return False

    @property
    def ordered_holding(self) -> list[list[int]]:
        """Returns suits and holdings in decreasing order of holding
            e.g. [[5, 1], [4, 0], [3, 2], [1, 3]].
        """
        holding = ([[self._spades, self.spade_suit],
                    [self._hearts, self.heart_suit],
                    [self._diamonds, self.diamond_suit],
                    [self._clubs, self.club_suit]])
        holding.sort(reverse=True)
        return holding

    def long_suit(self, minimum_holding: int) -> Suit | None:
        """Return the first suit with minimum_holding number of cards."""
        suit = None
        if self.spades >= minimum_holding:
            suit = self.spade_suit
        elif self.hearts >= minimum_holding:
            suit = self.heart_suit
        elif self.diamonds >= minimum_holding:
            suit = self.diamond_suit
        elif self.clubs >= minimum_holding:
            suit = self.club_suit
        return suit

    @property
    def suit_shape(self) -> list[int]:
        """Return a list of suits in decreasing order by holding."""
        shape = []
        for suit in self.ordered_holding:
            shape.append(suit[1])
        return shape

    @staticmethod
    def higher_ranking_suit(suit_one: Suit, suit_two: Suit) -> Suit:
        """Return the higher ranking of two suits."""
        if suit_one > suit_two:
            return suit_one
        return suit_two

    def has_sequence(self, suit) -> bool:
        """Return True if hand contains a three card sequence starting with
            an honour."""
        sequences = ['AKQ', 'KQJ', 'QJT', 'JT9', 'T98']
        cards = [card for card in self.cards if card.suit == suit]
        if len(cards) >= 3:
            for index, card in enumerate(cards[2:]):
                triple = cards[index].rank + cards[index+1].rank + card.rank
                if triple in sequences:
                    return True
        return False

    def suit_length(self, suit: Suit) -> int:
        """Return the length of a suit or a dummy value."""
        if suit.is_suit:
            return self.suit_holding[suit]
        return -999

    @staticmethod
    def _call_name(level, suit: Suit) -> str:
        """Return a call name form level and suit."""
        call_name = ''.join([str(level), suit])
        return call_name

    def double_allowed(self):
        bid_history = (['P', 'P', 'P'] + self.bid_history)[-3:]
        if bid_history[-1] == 'D':
            return False
        if (bid_history[-3] == 'P' and
                bid_history[-2] == 'P' and
                bid_history[-1] == 'P'):
            return False
        if (bid_history[-3] != 'P' and
                bid_history[-2] == 'P' and
                bid_history[-1] == 'P'):
            return True
        if (bid_history[-3] == 'P' and
                bid_history[-2] != 'P' and
                bid_history[-1] == 'P'):
            return False
        if (bid_history[-2] != 'P' and
                bid_history[-1] == 'P'):
            return False
        return True

    def redouble_allowed(self):
        bid_history = (['P', 'P', 'P'] + self.bid_history)[-3:]
        if bid_history[-1] == 'D':
            return True
        if (bid_history[-3] == 'D' and
                bid_history[-2] == 'P' and
                bid_history[-1]) == 'P':
            return True
        return False

    def tracer(self,
               module: str,
               get_frame: object,
               trace_value: str = '',
               display: bool = False,
               trace_message: str = '') -> None:
        """Print function trace information."""
        if self.display_hand:
            self._print_hand()

        if self.display_trace:
            self._print_trace(
                module, get_frame, trace_value, display, trace_message)

    def _print_hand(self) -> None:
        hand = self.__str__()
        if hand != self.last_hand:
            if self.last_hand:
                print('')
            cprint(
                f'{hand}, {self.hcp}, {self.shape} {self.hcp=}', MODULE_COLOUR)
            self._set_trace_hand()
            self.last_hand = hand

    def _print_trace(
            self,
            module: str,
            get_frame: object,
            trace_value: str = '',
            display: bool = False,
            trace_message: str = '') -> None:
        (source_file, display_file) = self._get_source_and_display(module)
        (call_id, call_name) = self._get_call_id_and_call_name(trace_value)
        line_number = get_frame.f_lineno
        function = get_frame.f_code.co_name
        if trace_message:
            trace_message = ''.join([', ', trace_message])

        output_list = [
            f'{line_number:03d}',
            f'{call_id=}',
            source_file,
            f'{function:.<40}',
            f'{call_name}{trace_message}'
        ]
        cprint(', '.join(output_list), TRACE_COLOUR)

    def force_trace(self) -> str | bool:
        try:
            with open(TRACE_PATH, 'r') as f_trace:
                trace_text = f_trace.read()
                if trace_text:
                    trace_value = True
                    return trace_value
        except FileNotFoundError:
            pass
        except NotADirectoryError:
            pass
        return False

    def _get_source_and_display(self, module: str) -> tuple[str, bool]:
        """Return a tuple (source_file, display)"""
        source_file = ''
        display = False
        if self.trace_module:
            source_file = module.replace('bfg_components.src.', '')
            if self.trace_module == source_file:
                display = True
        return (source_file, display)

    @staticmethod
    def _get_call_id_and_call_name(trace_value) -> tuple[str]:
        """Return a tuple (call_id, call_name)"""
        call_id = ''
        call_name = ''
        if isinstance(trace_value, Bid):
            call_id = trace_value.call_id
            call_name = trace_value.name
        return (call_id, call_name)

    def get_attributes_from_hand(self, hand):
        """Set the attributes of this object from a hand instance."""
        for key, item in hand.__dict__.items():
            self.__dict__[key] = item

    def _get_trace_module(self) -> tuple[str] | tuple[None]:
        """Return the name of the module to trace."""
        if os.path.isfile(MODULE_TRACE):
            with open(MODULE_TRACE, 'r') as f_trace_file:
                text = f_trace_file.read()
                text = text.split('\n')
                text.extend(['', ''])
                return (text[0], text[1])
        else:
            return (None, None)

    def _set_trace_hand(self) -> None:
        """Write the current hand to the trace file."""
        if os.path.isfile(MODULE_TRACE):
            with open(MODULE_TRACE, 'w') as f_trace_file:
                f_trace_file.write(
                    '\n'.join([self.trace_module, self.__str__()]))



""" Bid for Game
    Player class
"""
# from termcolor import cprint

from bridgeobjects import ROLES, Board, Hand
from .bidding import Bid
from .acol_bidding import AcolBid
from .utils import get_role, get_active_bid_history

MODULE_COLOUR = 'blue'


class Player(object):
    """Define BfG Player class."""
    NUMBER_OF_PLAYERS = 4

    def __init__(self,
                 board: Board | None = None,
                 hand: Hand | None = None,
                 index: int | None = None) -> None:
        self.board = board
        self.hand = hand
        self.index = index
        self.role = -1

    def __repr__(self) -> str:
        """Return a string representation of player."""
        return f'player: {self.hand}'

    def make_bid(self, update_bid_history: bool = True) -> Bid:
        """Make a bid and return bid object."""
        # roles = Roles(self.board.bid_history)
        self.role = get_role(self.board.bid_history)
        self.board.active_bid_history = get_active_bid_history(self.board.bid_history)

        bid = AcolBid(self.hand, self.board, self.role).bid

        if update_bid_history:
            self.board.bid_history.append(bid.name)
        hc_points = self.hand.high_card_points

        if bid.use_shortage_points:
            distribution_points = 0
            hand_points = f'{hc_points}+{distribution_points}'
            hand_description = f'{hand_points} = {hc_points+distribution_points}'
        else:
            hand_description = str(hc_points)
        hand_description = f'{hand_description} '
        bid.hand_points = hand_description
        return bid



"""Utility functions for bidding."""

from bridgeobjects import ROLES
from .bidding import Bid, Pass


NUMBER_OF_PLAYERS = 4


def get_role(bid_history: list[str]) -> int:
    """Return the role_id based on the length of the bid_history."""
    active_bid_history = get_active_bid_history(bid_history)
    if len(active_bid_history) % NUMBER_OF_PLAYERS == 0:
        return ROLES['Opener']

    if len(active_bid_history) % NUMBER_OF_PLAYERS == 2:
        return ROLES['Responder']

    all_passes = True
    for index, bid in enumerate(active_bid_history[1::2]):
        if bid != 'P':
            all_passes = False
            break

    if all_passes:  # Must be overcaller
        return ROLES['Overcaller']

    # Get number of calls after first overcall
    overcall_index = (index * 2) + 2
    remaining_calls = len(active_bid_history) - overcall_index + 1

    if remaining_calls % NUMBER_OF_PLAYERS == 0:
        return ROLES['Overcaller']
    return ROLES['Advancer']


def get_active_bid_history(bid_history) -> list[str]:
    """Return the bid history without leading PASSES."""
    for index, bid in enumerate(bid_history):
        if bid != 'P':
            return bid_history[index:]
    return []


class NamedBids():
    """Specifically named bids."""
    def __init__(self, hand: object, bid_history: list[str]) -> None:
        self.hand = hand
        self.bid_history = bid_history

        self.pass_bid = Pass('9999')
        self.opener_bid_one = self.pass_bid
        self.opener_bid_two = self.pass_bid
        self.opener_bid_three = self.pass_bid
        self.responder_bid_one = self.pass_bid
        self.responder_bid_two = self.pass_bid
        self.responder_bid_three = self.pass_bid
        self.overcaller_bid_one = self.pass_bid
        self.overcaller_bid_two = self.pass_bid
        self.overcallers_last_bid = self.pass_bid
        self.advancer_bid_one = self.pass_bid
        self.advancer_bid_two = self.pass_bid

        self.previous_bid = self.pass_bid
        self.last_bid = self.pass_bid
        self.right_hand_bid = self.pass_bid
        self.my_last_bid = self.pass_bid

        self.partner_bid_one = self.pass_bid
        self.partner_bid_two = self.pass_bid
        self.partner_penultimate_bid = self.pass_bid
        self.partner_last_bid = self.pass_bid

        self.bid_one = self.pass_bid
        self.bid_two = self.pass_bid
        self.bid_three = self.pass_bid

        self.responders_support = False
        self.overcaller_has_jumped = False
        self.bid_after_stayman = False
        self.overcaller_in_second_seat = self._overcaller_in_second_seat()
        self.overcaller_in_fourth_seat = self._overcaller_in_fourth_seat()

        self.assign_bids()

        self.suit_one = None
        if self.bid_one and self.bid_one.is_suit_call:
            self.suit_one = self.bid_one.denomination

        self.overcaller_suit_one = None
        self.advancer_suit_one = None
        self.advancer_suit_two = None
        self.holding_partner_one = 0
        if self.hand:
            if self.overcaller_bid_one.is_suit_call:
                self.overcaller_suit_one =  self.overcaller_bid_one.denomination
            if self.advancer_bid_one.is_suit_call:
                self.advancer_suit_one = self.advancer_bid_one.denomination
                self.holding_partner_one = self.hand.suit_length(self.advancer_suit_one)
            if self.advancer_bid_two.is_suit_call:
                self.advancer_suit_two = self.advancer_bid_two.denomination

    @staticmethod
    def is_jump(first_bid: Bid, second_bid: Bid) -> bool:
        """Return True if second bid is a jump over first."""
        # e.g.  first_bid = Bid('1S')
        #       second_bid = Bid('2NT')
        if second_bid.is_value_call and first_bid.is_value_call:
            jump_level = second_bid.level - first_bid.level
            if jump_level > 1:
                return True
            elif jump_level == 1:
                if second_bid.is_suit_call and first_bid.is_suit_call:
                    if second_bid.denomination > first_bid.denomination:
                        return True
                elif second_bid.is_nt:
                    return True
        return False

    def _bid_after_stayman(self) -> bool:
        """Return True if responder has bid Clubs after NT opening."""
        if (self.hand and self.opener_bid_one.is_nt and
                self.responder_bid_one.denomination == self.hand.club_suit):
            return True
        return False

    def _overcaller_in_second_seat(self) -> bool:
        """Assign seat to overcaller."""
        if (len(self.bid_history) >= 2 and
                Bid(self.bid_history[1]).name != Pass().name):
            return True
        return False

    def _overcaller_in_fourth_seat(self) -> bool:
        """Assign seat to overcaller."""
        if (len(self.bid_history) >= 2 and
                Bid(self.bid_history[1]).is_pass and
                len(self.bid_history) >= 4 and
                not Bid(self.bid_history[3]).is_pass):
            return True
        return False

    def assign_bids(self) -> None:
        """Assign bids to named attributes based on bid_history."""
        if not self.bid_history:
            return

        bids = len(self.bid_history)
        if bids >= 1:
            self.last_bid = Bid(self.bid_history[-1])
            self._assign_first_bid()
            if bids == 1:
                return
        if bids >= 2:
            self._assign_second_bid()
            if bids == 2:
                return
        if bids >= 3:
            self._assign_third_bid()
            if bids == 3:
                return
        if bids >= 4:
            self._assign_fourth_bid()
            if bids == 4:
                return
        if bids >= 5:
            self._assign_fifth_bid()
            if bids == 5:
                return
        if bids >= 6:
            self._assign_sixth_bid()
            if bids == 6:
                return
        if bids >= 7:
            self._assign_seventh_bid()
            if bids == 7:
                return
        if bids >= 8:
            self._assign_eighth_bid()
            if bids == 8:
                return
        if bids >= 9:
            self._assign_ninth_bid()
            if bids == 9:
                return
        if bids >= 10:
            self._assign_tenth_bid()
            if bids == 10:
                return
        if bids >= 11:
            self._assign_eleventh_bid()
            if bids == 11:
                return
        if bids >= 12:
            self._assign_twelfth_bid()
            if bids == 12:
                return

    def _assign_first_bid(self) -> None:
        """Assign first bid."""
        self.opener_bid_one = Bid(self.bid_history[0])

    def _assign_second_bid(self) -> None:
        """Assign second bid."""
        # if self.bid_history[1] == 'P':
        self.overcaller_bid_one = Bid(self.bid_history[1])
        self.overcaller_has_jumped = self.is_jump(self.opener_bid_one,
                                                self.overcaller_bid_one)
        self.right_hand_bid = Bid(self.bid_history[-1])
        self.partner_last_bid = Bid(self.bid_history[-2])

        if self.partner_bid_one == self.pass_bid:
            self.partner_bid_one = Bid(self.bid_history[-2])

        if self.hand and self.partner_bid_one.is_suit_call:
            self.holding_partner_one = self.hand.suit_holding[self.partner_bid_one.denomination]

        if len(self.bid_history) >= 7:
            self.partner_penultimate_bid = Bid(self.bid_history[-6])

    def _assign_third_bid(self) -> None:
        """Assign third bid."""
        self.responder_bid_one = Bid(self.bid_history[2])
        self.responders_support = (self.opener_bid_one.denomination ==
                                   self.responder_bid_one.denomination)

    def _assign_fourth_bid(self) -> None:
        """Assign fourth bid."""
        self.advancer_bid_one = Bid(self.bid_history[3])
        self.bid_one = Bid(self.bid_history[-4])
        self.previous_bid = Bid(self.bid_history[-4])
        if self.overcaller_in_fourth_seat:
            self.overcaller_bid_one = Bid(self.bid_history[3])
            self.overcaller_has_jumped = self.is_jump(self.responder_bid_one,
                                                      self.overcaller_bid_one)
        self.my_last_bid = Bid(self.bid_history[-4])

    def _assign_fifth_bid(self) -> None:
        """Assign fifth bid."""
        self.opener_bid_two = Bid(self.bid_history[4])
        self.bid_after_stayman = self._bid_after_stayman()

    def _assign_sixth_bid(self) -> None:
        """Assign sixth bid."""
        self.partner_bid_one = Bid(self.bid_history[-6])
        self.partner_bid_two = Bid(self.bid_history[-2])
        if self.overcaller_in_fourth_seat:
            self.overcaller_bid_one = Bid(self.bid_history[3])
            self.advancer_bid_one = Bid(self.bid_history[5])

    def _assign_seventh_bid(self) -> None:
        """Assign seventh bid."""
        self.responder_bid_two = Bid(self.bid_history[6])
        self.overcaller_bid_two = Bid(self.bid_history[5])

    def _assign_eighth_bid(self) -> None:
        """Assign eighth bid."""
        self.bid_one = Bid(self.bid_history[-8])
        self.bid_two = Bid(self.bid_history[-4])
        self.overcallers_last_bid = Bid(self.bid_history[-1])
        if self.overcaller_in_fourth_seat:
            self.overcaller_bid_two = Bid(self.bid_history[7])
        else:
            self.advancer_bid_two = Bid(self.bid_history[7])

    def _assign_ninth_bid(self) -> None:
        """Assign ninth bid."""
        self.opener_bid_three = Bid(self.bid_history[8], '')

    def _assign_tenth_bid(self) -> None:
        """Assign tenth bid."""
        self.partner_bid_one = Bid(self.bid_history[-10])
        self.partner_bid_two = Bid(self.bid_history[-6])

    def _assign_eleventh_bid(self) -> None:
        """Assign eleventh bid."""
        self.responder_bid_three = Bid(self.bid_history[10])
        # self.overcaller_bid_two = Bid(self.bid_history[5])

    def _assign_twelfth_bid(self) -> None:
        """Assign twelfth bid."""
        self.bid_one = Bid(self.bid_history[-12])
        self.bid_two = Bid(self.bid_history[-8])
        self.bid_three = Bid(self.bid_history[-4])
